/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"
#include "menu_common.h"


/*
=======================================================================

KEYS MENU

=======================================================================
*/
typedef struct keymenuitem_s
{
	char *command;
	char *text;
	menuaction_s action;
} keymenuitem_t;


keymenuitem_t key_menu_items[] = {
	{"+attack", "attack"},
	{"weapprev", "prev weapon"},
	{"weapnext", "next weapon"},
	{"+forward", "walk forward"},
	{"+back", "backpedal"},
	{"+left", "turn left"},
	{"+right", "turn right"},
	{"+speed", "run"},
	{"+moveleft", "step left"},
	{"+moveright", "step right"},
	{"+strafe", "sidestep"},
	{"+moveup", "up/jump"},
	{"+movedown", "down/crouch"},
	{"inven", "inventory"},
	{"invuse", "use item"},
	{"invdrop", "drop item"},
	{"invprev", "prev item"},
	{"invnext", "next item"},
	{"cmd help", "help computer"},
	{0, 0}
};


static int bind_grab;
static menuframework_s s_keys_menu;


static void M_UnbindCommand (char *command)
{
	int l = strlen (command);

	for (int j = 0; j < 256; j++)
	{
		char *b = keybindings[j];

		if (!b) continue;
		if (!strncmp (b, command, l)) Key_SetBinding (j, "");
	}
}


static void M_FindKeysForCommand (char *command, int *twokeys)
{
	twokeys[0] = twokeys[1] = -1;

	int l = strlen (command);
	int count = 0;

	for (int j = 0; j < 256; j++)
	{
		char *b = keybindings[j];

		if (!b)
			continue;

		if (!strncmp (b, command, l))
		{
			twokeys[count] = j;
			count++;

			if (count == 2)
				break;
		}
	}
}


static void KeyCursorDrawFunc (menuframework_s *menu)
{
	if (bind_grab)
		re->DrawChar (menu->x, menu->y + menu->cursor * 10 + 52, '=');
	else M_DrawPointer (menu->x, menu->y + menu->cursor * 10 + 52);
}


static void DrawKeyBindingFunc (void *self)
{
	int keys[2];
	menuaction_s *a = (menuaction_s *) self;

	M_FindKeysForCommand (key_menu_items[a->generic.localdata[0]].command, keys);

	if (keys[0] == -1)
		Menu_DrawString (a->generic.x + a->generic.parent->x + 16, a->generic.y + a->generic.parent->y, "???");
	else
	{
		int x;
		const char *name;

		name = Key_KeynumToString (keys[0]);

		Menu_DrawString (a->generic.x + a->generic.parent->x + 16, a->generic.y + a->generic.parent->y, name);

		x = strlen (name) * 8;

		if (keys[1] != -1)
		{
			Menu_DrawStringDark (a->generic.x + a->generic.parent->x + 24 + x, a->generic.y + a->generic.parent->y, "or");
			Menu_DrawString (a->generic.x + a->generic.parent->x + 48 + x, a->generic.y + a->generic.parent->y, Key_KeynumToString (keys[1]));
		}
	}
}

static void KeyBindingFunc (void *self)
{
	menuaction_s *a = (menuaction_s *) self;
	int keys[2];

	M_FindKeysForCommand (key_menu_items[a->generic.localdata[0]].command, keys);

	if (keys[1] != -1)
		M_UnbindCommand (key_menu_items[a->generic.localdata[0]].command);

	bind_grab = true;

	Menu_SetStatusBar (&s_keys_menu, "press a key or button for this action");
}


static void Keys_MenuInit (void)
{
	int y = 0;
	int i = 0;

	s_keys_menu.x = viddef.width / 2;
	s_keys_menu.y = viddef.height / 2 - 58;
	s_keys_menu.nitems = 0;
	s_keys_menu.cursordraw = KeyCursorDrawFunc;

	for (int i = 0, y = 52;; i++, y += 10)
	{
		// end of list
		if (!key_menu_items[i].command) break;
		if (!key_menu_items[i].text) break;

		key_menu_items[i].action.generic.type = MTYPE_ACTION;
		key_menu_items[i].action.generic.flags = QMF_GRAYED;
		key_menu_items[i].action.generic.x = 0;
		key_menu_items[i].action.generic.y = y;
		key_menu_items[i].action.generic.ownerdraw = DrawKeyBindingFunc;
		key_menu_items[i].action.generic.localdata[0] = i;
		key_menu_items[i].action.generic.name = key_menu_items[i].text;

		Menu_AddItem (&s_keys_menu, (void *) &key_menu_items[i].action);
	}

	Menu_SetStatusBar (&s_keys_menu, "enter to change, backspace to clear");
	Menu_Center (&s_keys_menu);
}


static void Keys_MenuDraw (void)
{
	M_Banner ("m_banner_customize");
	Menu_AdjustCursor (&s_keys_menu, 1);
	Menu_Draw (&s_keys_menu);
}


static const char *Keys_MenuKey (int key)
{
	menuaction_s *item = (menuaction_s *) Menu_ItemAtCursor (&s_keys_menu);

	if (bind_grab)
	{
		if (key != K_ESCAPE && key != '`')
		{
			char cmd[1024];

			Com_sprintf (cmd, sizeof (cmd), "bind \"%s\" \"%s\"\n", Key_KeynumToString (key), key_menu_items[item->generic.localdata[0]].command);
			Cbuf_InsertText (cmd);
		}

		Menu_SetStatusBar (&s_keys_menu, "enter to change, backspace to clear");
		bind_grab = false;
		return menu_out_sound;
	}

	switch (key)
	{
	case K_KP_ENTER:
	case K_ENTER:
		KeyBindingFunc (item);
		return menu_in_sound;

	case K_BACKSPACE:		// delete bindings
	case K_DEL:				// delete bindings
	case K_KP_DEL:
		M_UnbindCommand (key_menu_items[item->generic.localdata[0]].command);
		return menu_out_sound;

	default:
		return Default_MenuKey (&s_keys_menu, key);
	}
}

void M_Menu_Keys_f (void)
{
	Keys_MenuInit ();
	M_PushMenu (Keys_MenuDraw, Keys_MenuKey);
}


