/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"
#include "winquake.h"
#include "menu_common.h"

// needed for PrepVideoMenu
#include <d3d9.h>

extern void M_ForceMenuOff (void);

void VID_MenuInit (void);

/*
====================================================================

MENU INTERACTION

====================================================================
*/

static menuframework_s	s_vid_menu[2];

static menulist_s		s_mode_list;
static menunumberlist_s	s_width_list;
static menunumberlist_s	s_height_list;
static menuslider_s		s_gamma_slider;
static menuslider_s		s_intensity_slider;
static menuslider_s		s_brightness_slider;
static menulist_s  		s_fs_box;
static menulist_s  		s_finish_box;
static menuaction_s		s_apply_action;
static menuaction_s		s_defaults_action;

// we fixed the max number of modes to 1024 so this is safe
static itemlist_t fs_resolutions;

static int win_widths[1024];
static int win_heights[1024];
static int win_numwidths;
static int win_numheights;

int M_ModeSortFunc (int *a, int *b)
{
	return *a - *b;
}


static void FullscreenCallback (void *s)
{
	// sync the cursors (assumes that fs is option 0)
	s_vid_menu[0].cursor = s_vid_menu[1].cursor = 0;
}


static void GammaCallback (void *s)
{
	menuslider_s *slider = (menuslider_s *) s;
	Cvar_SetValue ("vid_gamma", (0.8 - (slider->curvalue / 10.0 - 0.5)) + 0.5);
}

static void IntensityCallback (void *s)
{
	menuslider_s *slider = (menuslider_s *) s;
	Cvar_SetValue ("intensity", slider->curvalue / 3.0f);
}

static void BrightnessCallback (void *s)
{
	menuslider_s *slider = (menuslider_s *) s;
	Cvar_SetValue ("brightness", slider->curvalue / 6.0f);
}

static void ResetDefaults (void *unused)
{
	// re-init the video menu to reload current values
	VID_MenuInit ();
}

static void ApplyChanges (void *unused)
{
	// running the callbacks to keep everything consistent
	GammaCallback (&s_gamma_slider);
	IntensityCallback (&s_intensity_slider);
	BrightnessCallback (&s_brightness_slider);

	Cvar_SetValue ("vid_fullscreen", s_fs_box.curvalue);
	Cvar_SetValue ("gl_swapinterval", s_finish_box.curvalue);

	if (s_fs_box.curvalue)
		Cvar_SetValue ("vid_mode", s_mode_list.curvalue);
	else
	{
		// using Cvar_Set instead of Cvar_SetValue because these are stored as strings for the menu
		Cvar_SetValue ("vid_width", win_widths[s_width_list.curvalue]);
		Cvar_SetValue ("vid_height", win_heights[s_height_list.curvalue]);
	}

	M_ForceMenuOff ();
}

static void CancelChanges (void *unused)
{
	extern void M_PopMenu (void);

	M_PopMenu ();
}


qboolean VID_MenuFindValue (int *values, int numvalues, int val)
{
	for (int i = 0; i < numvalues; i++)
	{
		if (values[i] == val)
			return true;
	}

	return false;
}


/*
===================
VID_MenuInit
===================
*/
void VID_MenuInit (void)
{
	static const char *yesno_names[] = {
		"no",
		"yes",
		0
	};

	s_mode_list.curvalue = Cvar_VariableValue ("vid_mode");

	// add the current dimensions to the windowed mode list if necessary and update the lists
	RECT crect;

	GetClientRect (cl_hwnd, &crect);

	int cw = crect.right - crect.left;
	int ch = crect.bottom - crect.top;

	if (!VID_MenuFindValue (win_widths, win_numwidths, cw))
	{
		if (win_numwidths < 1024)
		{
			win_widths[win_numwidths] = cw;
			win_numwidths++;
			qsort (win_widths, win_numwidths, sizeof (int), (sortfunc_t) M_ModeSortFunc);
		}
	}

	if (!VID_MenuFindValue (win_heights, win_numheights, ch))
	{
		if (win_numheights < 1024)
		{
			win_heights[win_numheights] = ch;
			win_numheights++;
			qsort (win_heights, win_numheights, sizeof (int), (sortfunc_t) M_ModeSortFunc);
		}
	}

	// init these to safe values
	s_width_list.curvalue = 0;
	s_height_list.curvalue = 0;

	// get correct width value
	for (int i = 0; i < win_numwidths; i++)
	{
		if (win_widths[i] >= cw)
		{
			s_width_list.curvalue = i;
			break;
		}
	}

	// get correct height value
	for (int i = 0; i < win_numheights; i++)
	{
		if (win_heights[i] >= ch)
		{
			s_height_list.curvalue = i;
			break;
		}
	}

	for (int i = 0; i < 2; i++)
	{
		s_vid_menu[i].x = viddef.width * 0.50;
		s_vid_menu[i].nitems = 0;

		s_fs_box.generic.type = MTYPE_SPINCONTROL;
		s_fs_box.generic.x = 0;
		s_fs_box.generic.y = 0;
		s_fs_box.generic.callback = FullscreenCallback;
		s_fs_box.generic.name = "fullscreen";
		s_fs_box.itemnames = yesno_names;
		s_fs_box.curvalue = Cvar_VariableValue ("vid_fullscreen");
		Menu_AddItem (&s_vid_menu[i], (void *) &s_fs_box);

		if (i == 0)
		{
			// fullscreen modes - mode and refresh rate
			s_mode_list.generic.type = MTYPE_SPINCONTROL;
			s_mode_list.generic.name = "video mode";
			s_mode_list.generic.x = 0;
			s_mode_list.generic.y = 10;
			s_mode_list.itemnames = fs_resolutions.items;
			Menu_AddItem (&s_vid_menu[i], (void *) &s_mode_list);
		}
		else
		{
			// windowed modes - height and width
			s_width_list.generic.type = MTYPE_NUMBERLIST;
			s_width_list.generic.name = "width";
			s_width_list.generic.x = 0;
			s_width_list.generic.y = 10;
			s_width_list.values = win_widths;
			s_width_list.numvalues = win_numwidths;
			Menu_AddItem (&s_vid_menu[i], (void *) &s_width_list);

			s_height_list.generic.type = MTYPE_NUMBERLIST;
			s_height_list.generic.name = "height";
			s_height_list.generic.x = 0;
			s_height_list.generic.y = 20;
			s_height_list.values = win_heights;
			s_height_list.numvalues = win_numheights;
			Menu_AddItem (&s_vid_menu[i], (void *) &s_height_list);
		}

		s_finish_box.generic.type = MTYPE_SPINCONTROL;
		s_finish_box.generic.x = 0;
		s_finish_box.generic.y = 30;
		s_finish_box.generic.name = "vertical sync";
		s_finish_box.curvalue = Cvar_VariableValue ("gl_swapinterval");
		s_finish_box.itemnames = yesno_names;
		Menu_AddItem (&s_vid_menu[i], (void *) &s_finish_box);

		s_gamma_slider.generic.type = MTYPE_SLIDER;
		s_gamma_slider.generic.x = 0;
		s_gamma_slider.generic.y = 50;
		s_gamma_slider.generic.name = "gamma";
		s_gamma_slider.generic.callback = GammaCallback;
		s_gamma_slider.minvalue = 5;
		s_gamma_slider.maxvalue = 13;
		s_gamma_slider.curvalue = (1.3 - Cvar_VariableValue ("vid_gamma") + 0.5) * 10;
		Menu_AddItem (&s_vid_menu[i], (void *) &s_gamma_slider);

		s_intensity_slider.generic.type = MTYPE_SLIDER;
		s_intensity_slider.generic.x = 0;
		s_intensity_slider.generic.y = 60;
		s_intensity_slider.generic.name = "intensity";
		s_intensity_slider.generic.callback = IntensityCallback;
		s_intensity_slider.minvalue = 0;
		s_intensity_slider.maxvalue = 12;
		s_intensity_slider.curvalue = Cvar_VariableValue ("intensity") * 3;
		Menu_AddItem (&s_vid_menu[i], (void *) &s_intensity_slider);

		s_brightness_slider.generic.type = MTYPE_SLIDER;
		s_brightness_slider.generic.x = 0;
		s_brightness_slider.generic.y = 70;
		s_brightness_slider.generic.name = "brightness";
		s_brightness_slider.generic.callback = BrightnessCallback;
		s_brightness_slider.minvalue = 0;
		s_brightness_slider.maxvalue = 12;
		s_brightness_slider.curvalue = Cvar_VariableValue ("brightness") * 6;
		Menu_AddItem (&s_vid_menu[i], (void *) &s_brightness_slider);

		s_defaults_action.generic.type = MTYPE_ACTION;
		s_defaults_action.generic.name = "reset to defaults";
		s_defaults_action.generic.x = 0;
		s_defaults_action.generic.y = 90;
		s_defaults_action.generic.callback = ResetDefaults;
		Menu_AddItem (&s_vid_menu[i], (void *) &s_defaults_action);

		s_apply_action.generic.type = MTYPE_ACTION;
		s_apply_action.generic.name = "apply";
		s_apply_action.generic.x = 0;
		s_apply_action.generic.y = 100;
		s_apply_action.generic.callback = ApplyChanges;
		Menu_AddItem (&s_vid_menu[i], (void *) &s_apply_action);

		Menu_Center (&s_vid_menu[i]);
		s_vid_menu[i].x -= 8;
	}
}

/*
================
VID_MenuDraw
================
*/
void VID_MenuDraw (void)
{
	// draw the banner
	M_Banner ("m_banner_video");

	if (s_fs_box.curvalue)
	{
		Menu_AdjustCursor (&s_vid_menu[0], 1);
		Menu_Draw (&s_vid_menu[0]);
	}
	else
	{
		Menu_AdjustCursor (&s_vid_menu[1], 1);
		Menu_Draw (&s_vid_menu[1]);
	}
}

/*
================
VID_MenuKey
================
*/
const char *VID_MenuKey (int key)
{
	static const char *sound = "misc/menu1.wav";
	menuframework_s *currentmenu = s_fs_box.curvalue ? &s_vid_menu[0] : &s_vid_menu[1];

	switch (key)
	{
	case K_ESCAPE:
		CancelChanges (NULL);
		return NULL;

	case K_KP_UPARROW:
	case K_UPARROW:
		currentmenu->cursor--;
		Menu_AdjustCursor (currentmenu, -1);
		break;

	case K_KP_DOWNARROW:
	case K_DOWNARROW:
		currentmenu->cursor++;
		Menu_AdjustCursor (currentmenu, 1);
		break;

	case K_KP_LEFTARROW:
	case K_LEFTARROW:
		Menu_SlideItem (currentmenu, -1);
		break;

	case K_KP_RIGHTARROW:
	case K_RIGHTARROW:
		Menu_SlideItem (currentmenu, 1);
		break;

	case K_KP_ENTER:
	case K_ENTER:
		Menu_SelectItem (currentmenu);
		break;
	}

	return sound;
}


void M_Menu_Video_f (void)
{
	VID_MenuInit ();
	M_PushMenu (VID_MenuDraw, VID_MenuKey);
}


typedef struct resolutions_s
{
	D3DDISPLAYMODE *modes;
	int nummodes;
} resolutions_t;


void M_BuildFSResolutionList (itemlist_t *list, void *extradata)
{
	resolutions_t *reslist = (resolutions_t *) extradata;

	// build the fs mode list
	for (int i = 0; i < reslist->nummodes; i++)
	{
		char modetemp[256];

		sprintf (modetemp, "%ix%i\n(%iHz)", reslist->modes[i].Width, reslist->modes[i].Height, reslist->modes[i].RefreshRate);
		CopyItemListItem (list, modetemp);
	}
}


void M_PrepVideoMenu (D3DDISPLAYMODE *modes, int nummodes)
{
	// max we'll display in the menus (the 1 over is for null-termination)
	if (nummodes > 1023) nummodes = 1023;

	resolutions_t reslist = {modes, nummodes};

	// build the fs mode list
	BuildItemList (&fs_resolutions, M_BuildFSResolutionList, &reslist, false);

	// init counts
	win_numwidths = 0;
	win_numheights = 0;

	// build the lists
	// no need to check for win_num* < 1024 here because we clamped nummodes above
	for (int i = 0; i < nummodes; i++)
	{
		if (!VID_MenuFindValue (win_widths, win_numwidths, modes[i].Width))
		{
			win_widths[win_numwidths] = modes[i].Width;
			win_numwidths++;
		}

		if (!VID_MenuFindValue (win_heights, win_numheights, modes[i].Height))
		{
			win_heights[win_numheights] = modes[i].Height;
			win_numheights++;
		}
	}

	// sort the lists
	qsort (win_widths, win_numwidths, sizeof (int), (sortfunc_t) M_ModeSortFunc);
	qsort (win_heights, win_numheights, sizeof (int), (sortfunc_t) M_ModeSortFunc);

	// precache video menu initialization (??? so that it won't falsely detect a mode change the first time we go into it ???)
	VID_MenuInit ();
}

