/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "gl_local.h"
#include "resource.h"

IDirect3DVertexShader9 *d3d_BeamVS = NULL;
IDirect3DPixelShader9 *d3d_BeamPS = NULL;
IDirect3DVertexDeclaration9 *d3d_BeamDecl = NULL;
IDirect3DVertexBuffer9 *d3d_BeamVertexes = NULL;

int r_numbeamverts = 0;
cvar_t *r_beamdetail = NULL;

void Beam_OnCreate (void)
{
	D3DVERTEXELEMENT9 layout[] = {
		VDECL (0, 0, D3DDECLTYPE_FLOAT3, D3DDECLUSAGE_POSITION, 0),
		D3DDECL_END ()
	};

	d3d_BeamDecl = R_CreateVertexDeclaration (layout);

	R_CreateShaders (BeamVS, BeamPS, IDR_BEAMSHADER);
}

void Beam_OnLost (void)
{
	SAFE_RELEASE (d3d_BeamVertexes);
}

void Beam_OnReset (void)
{
	if (!r_beamdetail) r_beamdetail = ri->Cvar_Get ("r_beamdetail", "60", CVAR_ARCHIVE);

	int slices = r_beamdetail->value;
	float *verts = NULL;

	// clamp sensibly
	if (slices < 3) slices = 3;
	if (slices > 360) slices = 360;

	r_numbeamverts = (slices + 1) * 2;
	d3d_Device->lpVtbl->CreateVertexBuffer (d3d_Device, r_numbeamverts * sizeof (float) * 3, D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &d3d_BeamVertexes, NULL);

	if (SUCCEEDED (d3d_BeamVertexes->lpVtbl->Lock (d3d_BeamVertexes, 0, 0, (void **) &verts, 0)))
	{
		for (int i = 0; i <= slices; i++, verts += 6)
		{
			float angle = (i == slices) ? 0.0f : (2.0f * M_PI * (float) i / (float) slices);

			Vector3Set (&verts[0], sin (angle) * 0.5f, cos (angle) * 0.5f, 1);
			Vector3Set (&verts[3], sin (angle) * 0.5f, cos (angle) * 0.5f, 0);
		}

		d3d_BeamVertexes->lpVtbl->Unlock (d3d_BeamVertexes);
	}
	else Beam_OnLost ();

	// don't trigger unnecessarily
	r_beamdetail->modified = false;
}

void Beam_OnDestroy (void) {}

void Beam_Register (void)
{
	static qboolean registered = false;

	if (!registered)
	{
		R_RegisterHandler (Beam_OnCreate, Beam_OnLost, Beam_OnReset, Beam_OnDestroy);
		registered = true;
	}
}


/*
===================
R_DrawBeam
===================
*/
void R_DrawBeam (entity_t *e)
{
	if (!r_beamdetail) r_beamdetail = ri->Cvar_Get ("r_beamdetail", "60", CVAR_ARCHIVE);

	if (r_beamdetail->modified)
	{
		Beam_OnLost ();
		Beam_OnReset ();
		r_beamdetail->modified = false;
	}

	if (r_numbeamverts < 1) return;

	float mins[3] = {999999, 999999, 999999};
	float maxs[3] = {-999999, -999999, -999999};

	for (int i = 0; i < 3; i++)
	{
		if (e->currorigin[i] < mins[i]) mins[i] = e->currorigin[i];
		if (e->currorigin[i] > maxs[i]) maxs[i] = e->currorigin[i];
		if (e->prevorigin[i] < mins[i]) mins[i] = e->prevorigin[i];
		if (e->prevorigin[i] > maxs[i]) maxs[i] = e->prevorigin[i];
	}

	// currframe is a radius, so halve it (rounding up in case it's odd)
	// we don't know the full orientation of the beam so just adjust all directions
	Vector3Subtractf (mins, mins, (e->currframe + 1) / 2);
	Vector3Addf (maxs, maxs, (e->currframe + 1) / 2);

	if (R_CullBox (mins, maxs)) return;

	float dir[3], len, axis[3];

	R_IdentityMatrix (&e->localmatrix);

	Vector3Subtract (dir, e->prevorigin, e->currorigin);
	Vector3Cross (axis, float3 (0, 0, 1), dir);

	// catch 0 length beams
	if ((len = Vector3Length (dir)) < 0.000001f) return;

	// and transform it
	R_TranslateMatrix (&e->localmatrix, e->currorigin[0], e->currorigin[1], e->currorigin[2]);
	R_RotateMatrixAxis (&e->localmatrix, (180.0f / M_PI) * acos ((Vector3Dot (float3 (0, 0, 1), dir) / len)), axis[0], axis[1], axis[2]);
	R_ScaleMatrix (&e->localmatrix, e->currframe, e->currframe, len);

	R_SetEntityTransform (&e->localmatrix, 0);

	// flip bgra to rgba
	R_SetPixelShaderConstant4f (
		PSC_SHADELIGHT,
		(float) ((byte *) &d_8to24table[e->skinnum & 0xff])[2] / 255.0f,
		(float) ((byte *) &d_8to24table[e->skinnum & 0xff])[1] / 255.0f,
		(float) ((byte *) &d_8to24table[e->skinnum & 0xff])[0] / 255.0f,
		e->alpha
	);

	R_SetVertexShader (d3d_BeamVS);
	R_SetPixelShader (d3d_BeamPS);
	R_SetVertexDeclaration (d3d_BeamDecl);

	// set up trans flag
	if (e->alpha < 1)
		e->flags |= RF_TRANSLUCENT;
	else e->flags &= ~RF_TRANSLUCENT;

	// and set the state
	R_SetEntityState (e);
	R_SetVertexBuffer (0, d3d_BeamVertexes, 0, sizeof (float) * 3);

	d3d_Device->lpVtbl->DrawPrimitive (d3d_Device, D3DPT_TRIANGLESTRIP, 0, r_numbeamverts - 2);
	c_draw_call++;
}


