/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "gl_local.h"


/*
=================================================================

PALETTE CONVERSION

=================================================================
*/

unsigned *R_Palette24To32 (const byte *pal24)
{
	static unsigned table32[256];

	for (int i = 0; i < 256; i++, pal24 += 3)
		table32[i] = LittleLong ((255 << 24) + (pal24[2] << 0) + (pal24[1] << 8) + (pal24[0] << 16));

	return table32;
}


/*
=================================================================

PCX LOADING

=================================================================
*/


/*
==============
LoadPCX
==============
*/
void R_LoadPCX (char *filename, byte **pic, byte **palette, int *width, int *height, void *(*allocator) (int))
{
	*pic = NULL;
	*palette = NULL;

	// load the file
	byte *raw = NULL;
	int len = ri->FS_LoadFile (filename, (void **) &raw);

	if (!raw)
	{
		ri->Con_Printf (PRINT_DEVELOPER, "Bad pcx file %s\n", filename);
		return;
	}

	// parse the PCX file
	pcx_t *pcx = (pcx_t *) raw;

	pcx->xmin = LittleShort (pcx->xmin);
	pcx->ymin = LittleShort (pcx->ymin);
	pcx->xmax = LittleShort (pcx->xmax);
	pcx->ymax = LittleShort (pcx->ymax);
	pcx->hres = LittleShort (pcx->hres);
	pcx->vres = LittleShort (pcx->vres);
	pcx->bytes_per_line = LittleShort (pcx->bytes_per_line);
	pcx->palette_type = LittleShort (pcx->palette_type);

	raw = &pcx->data;

	if (pcx->manufacturer != 0x0a || pcx->version != 5 || pcx->encoding != 1 || pcx->bits_per_pixel != 8 || pcx->xmax >= 640 || pcx->ymax >= 480)
	{
		ri->Con_Printf (PRINT_ALL, "Bad pcx file %s\n", filename);
		return;
	}

	byte *out = allocator ((pcx->ymax + 1) * (pcx->xmax + 1));
	byte *pix = out;

	*pic = out;

	if (palette)
	{
		*palette = allocator (768);
		memcpy (*palette, (byte *) pcx + len - 768, 768);
	}

	if (width) *width = pcx->xmax + 1;
	if (height) *height = pcx->ymax + 1;

	for (int y = 0; y <= pcx->ymax; y++, pix += pcx->xmax + 1)
	{
		for (int x = 0; x <= pcx->xmax;)
		{
			int dataByte = *raw++;
			int runLength = 1;

			if ((dataByte & 0xC0) == 0xC0)
			{
				runLength = dataByte & 0x3F;
				dataByte = *raw++;
			}

			while (runLength-- > 0)
				pix[x++] = dataByte;
		}
	}

	if (raw - (byte *) pcx > len)
	{
		ri->Con_Printf (PRINT_DEVELOPER, "PCX file %s was malformed", filename);
		*pic = NULL;
	}

	ri->FS_FreeFile (pcx);
}

/*
=========================================================

TARGA LOADING

=========================================================
*/

typedef struct _TargaHeader
{
	uint8_t 	id_length, colormap_type, image_type;
	uint16_t	colormap_index, colormap_length;
	uint8_t	colormap_size;
	uint16_t	x_origin, y_origin, width, height;
	uint8_t	pixel_size, attributes;
} TargaHeader;


/*
=============
LoadTGA
=============
*/
void R_LoadTGA (char *name, byte **pic, int *width, int *height, void *(*allocator) (int))
{
	int		columns, rows, numPixels;
	byte	*pixbuf;
	int		row, column;
	byte	*buf_p;
	byte	*buffer;
	int		length;
	TargaHeader		targa_header;
	byte			*targa_rgba;
	byte tmp[2];

	*pic = NULL;

	// load the file
	length = ri->FS_LoadFile (name, (void **) &buffer);

	if (!buffer)
	{
		ri->Con_Printf (PRINT_DEVELOPER, "Bad tga file %s\n", name);
		return;
	}

	buf_p = buffer;

	targa_header.id_length = *buf_p++;
	targa_header.colormap_type = *buf_p++;
	targa_header.image_type = *buf_p++;

	tmp[0] = buf_p[0];
	tmp[1] = buf_p[1];
	targa_header.colormap_index = LittleShort (*((short *) tmp));
	buf_p += 2;
	tmp[0] = buf_p[0];
	tmp[1] = buf_p[1];
	targa_header.colormap_length = LittleShort (*((short *) tmp));
	buf_p += 2;
	targa_header.colormap_size = *buf_p++;
	targa_header.x_origin = LittleShort (*((short *) buf_p));
	buf_p += 2;
	targa_header.y_origin = LittleShort (*((short *) buf_p));
	buf_p += 2;
	targa_header.width = LittleShort (*((short *) buf_p));
	buf_p += 2;
	targa_header.height = LittleShort (*((short *) buf_p));
	buf_p += 2;
	targa_header.pixel_size = *buf_p++;
	targa_header.attributes = *buf_p++;

	if (targa_header.image_type != 2 && targa_header.image_type != 10)
		ri->Sys_Error (ERR_DROP, "LoadTGA: Only type 2 and 10 targa RGB images supported\n");

	if (targa_header.colormap_type != 0 || (targa_header.pixel_size != 32 && targa_header.pixel_size != 24))
		ri->Sys_Error (ERR_DROP, "LoadTGA: Only 32 or 24 bit images supported (no colormaps)\n");

	columns = targa_header.width;
	rows = targa_header.height;
	numPixels = columns * rows;

	if (width) *width = columns;
	if (height) *height = rows;

	targa_rgba = allocator (numPixels * 4);
	*pic = targa_rgba;

	if (targa_header.id_length != 0)
		buf_p += targa_header.id_length;  // skip TARGA image comment

	if (targa_header.image_type == 2)  // Uncompressed, RGB images
	{
		for (row = rows - 1; row >= 0; row--)
		{
			pixbuf = targa_rgba + row * columns * 4;

			for (column = 0; column < columns; column++)
			{
				uint8_t red, green, blue, alphabyte;

				switch (targa_header.pixel_size)
				{
				case 24:
					red = *buf_p++;
					green = *buf_p++;
					blue = *buf_p++;
					*pixbuf++ = red;
					*pixbuf++ = green;
					*pixbuf++ = blue;
					*pixbuf++ = 255;
					break;

				case 32:
					red = *buf_p++;
					green = *buf_p++;
					blue = *buf_p++;
					alphabyte = *buf_p++;
					*pixbuf++ = red;
					*pixbuf++ = green;
					*pixbuf++ = blue;
					*pixbuf++ = alphabyte;
					break;
				}
			}
		}
	}
	else if (targa_header.image_type == 10)   // Runlength encoded RGB images
	{
		uint8_t red, green, blue, alphabyte, packetHeader, packetSize, j;

		for (row = rows - 1; row >= 0; row--)
		{
			pixbuf = targa_rgba + row * columns * 4;

			for (column = 0; column < columns;)
			{
				packetHeader = *buf_p++;
				packetSize = 1 + (packetHeader & 0x7f);

				if (packetHeader & 0x80)          // run-length packet
				{
					switch (targa_header.pixel_size)
					{
					case 24:
						red = *buf_p++;
						green = *buf_p++;
						blue = *buf_p++;
						alphabyte = 255;
						break;

					case 32:
						red = *buf_p++;
						green = *buf_p++;
						blue = *buf_p++;
						alphabyte = *buf_p++;
						break;
					}

					for (j = 0; j < packetSize; j++)
					{
						*pixbuf++ = red;
						*pixbuf++ = green;
						*pixbuf++ = blue;
						*pixbuf++ = alphabyte;
						column++;

						if (column == columns) // run spans across rows
						{
							column = 0;

							if (row > 0)
								row--;
							else goto breakOut;

							pixbuf = targa_rgba + row * columns * 4;
						}
					}
				}
				else                              // non run-length packet
				{
					for (j = 0; j < packetSize; j++)
					{
						switch (targa_header.pixel_size)
						{
						case 24:
							red = *buf_p++;
							green = *buf_p++;
							blue = *buf_p++;
							*pixbuf++ = red;
							*pixbuf++ = green;
							*pixbuf++ = blue;
							*pixbuf++ = 255;
							break;

						case 32:
							red = *buf_p++;
							green = *buf_p++;
							blue = *buf_p++;
							alphabyte = *buf_p++;
							*pixbuf++ = red;
							*pixbuf++ = green;
							*pixbuf++ = blue;
							*pixbuf++ = alphabyte;
							break;
						}

						column++;

						if (column == columns) // pixel packet run spans across rows
						{
							column = 0;

							if (row > 0)
								row--;
							else goto breakOut;

							pixbuf = targa_rgba + row * columns * 4;
						}
					}
				}
			}

breakOut:;
		}
	}

	ri->FS_FreeFile (buffer);
}


void R_LoadWAL (char *name, byte **pic, int *width, int *height, void *(*allocator) (int))
{
	miptex_t *mt = NULL;

	ri->FS_LoadFile (name, (void **) &mt);

	if (!mt)
	{
		ri->Con_Printf (PRINT_ALL, "R_FindImage: can't load %s\n", name);
		return;
	}

	*width = mt->width;
	*height = mt->height;

	*pic = allocator (mt->width * mt->height);

	memcpy (*pic, (byte *) mt + mt->offsets[0], mt->width * mt->height);

	ri->FS_FreeFile ((void *) mt);
}



