/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "gl_local.h"


/*
====================================================================

IMAGE FLOOD FILLING

====================================================================
*/


/*
=================
Mod_FloodFillSkin

Fill background pixels so mipmapping doesn't have haloes
=================
*/

typedef struct floodfill_s
{
	short x, y;
} floodfill_t;

// must be a power of 2
#define FLOODFILL_FIFO_SIZE 0x1000
#define FLOODFILL_FIFO_MASK (FLOODFILL_FIFO_SIZE - 1)

#define FLOODFILL_STEP(off, dx, dy) \
{ \
	if (pos[off] == fillcolor) \
	{ \
		pos[off] = 255; \
		fifo[inpt].x = x + (dx); \
		fifo[inpt].y = y + (dy); \
		inpt = (inpt + 1) & FLOODFILL_FIFO_MASK; \
	} \
	else if (pos[off] != 255) fdc = pos[off]; \
}


void R_FloodFillSkin (byte *skin, int skinwidth, int skinheight, unsigned *palette)
{
	byte		fillcolor = *skin; // assume this is the pixel to fill
	floodfill_t	*fifo = (floodfill_t *) ri->Scratch_Alloc ();
	int			inpt = 0, outpt = 0;
	int			filledcolor = -1;
	int			i;

	if (filledcolor == -1)
	{
		filledcolor = 0;

		// attempt to find opaque black
		for (i = 0; i < 256; ++i)
		{
			if (palette[i] == (255 << 0)) // alpha 1.0
			{
				filledcolor = i;
				break;
			}
		}
	}

	// can't fill to filled color or to transparent color (used as visited marker)
	if ((fillcolor != filledcolor) && (fillcolor != 255))
	{
		fifo[inpt].x = 0, fifo[inpt].y = 0;
		inpt = (inpt + 1) & FLOODFILL_FIFO_MASK;

		while (outpt != inpt)
		{
			int x = fifo[outpt].x, y = fifo[outpt].y;
			int fdc = filledcolor;
			byte *pos = &skin[x + skinwidth * y];

			outpt = (outpt + 1) & FLOODFILL_FIFO_MASK;

			if (x > 0)				FLOODFILL_STEP (-1, -1, 0);
			if (x < skinwidth - 1)	FLOODFILL_STEP (1, 1, 0);
			if (y > 0)				FLOODFILL_STEP (-skinwidth, 0, -1);
			if (y < skinheight - 1)	FLOODFILL_STEP (skinwidth, 0, 1);

			skin[x + skinwidth * y] = fdc;
		}
	}
}


void R_GenerateMiplevels (IDirect3DSurface9 *src, int dstlevel)
{
	if (!src) return;

	// generates miplevels for a full mipmap chain for the texture specified.  robust if only level 0 is present.
	// it is the caller's responsibility to get the underlying surface and clean up that surface when done.
	// recursive just because it's evil.
	IDirect3DTexture9 *tex = NULL;
	IDirect3DSurface9 *dst = NULL;

	if (SUCCEEDED (src->lpVtbl->GetContainer (src, &IID_IDirect3DTexture9, (void **) &tex)))
	{
		// this will just fail when it reaches the end of the mipmap chain and start dropping out of the recursion
		if (SUCCEEDED (tex->lpVtbl->GetSurfaceLevel (tex, dstlevel, &dst)))
		{
			// load this level
			R_LoadSurfaceFromSurface (dst, src);

			// recurse to the next miplevel
			R_GenerateMiplevels (dst, dstlevel + 1);

			// don't leak
			dst->lpVtbl->Release (dst);
			tex->lpVtbl->Release (tex);
		}
	}
}


void R_Upload8 (unsigned *dst, byte *src, int width, int height, int stride, unsigned *palette, qboolean alpha)
{
	for (int y = 0; y < height; y++)
	{
		for (int x = 0; x < width; x++)
		{
			if (alpha && src[x] == 255)
				dst[x] = 0;
			else dst[x] = palette[src[x]];
		}

		src += width;
		dst += stride;
	}
}


void R_Upload32 (unsigned *dst, unsigned *src, int width, int height, int stride)
{
	for (int y = 0; y < height; y++)
	{
		for (int x = 0; x < width; x++)
			dst[x] = src[x];

		src += width;
		dst += stride;
	}
}


byte *R_8to32 (byte *data, int width, int height, unsigned *palette, qboolean alpha)
{
	// because this doesn't advance the lowmark it's not safe to do a further allocation until after the data has been consumed
	unsigned *trans = (unsigned *) ri->Scratch_Alloc ();

	// send through R_Upload8 so that we get consistent alpha handling on both paths
	R_Upload8 (trans, data, width, height, width, palette, alpha);

	return (byte *) trans;
}


qboolean R_VerifyTextureAlpha (byte *data, int width, int height, unsigned *palette)
{
	if (palette)
	{
		for (int i = 0; i < width * height; i++)
			if (data[i] == 255)
				return true;
	}
	else
	{
		for (int i = 0; i < width * height; i++, data += 4)
			if (data[3] < 255)
				return true;
	}

	return false;
}


void R_FillTexture (IDirect3DTexture9 *tex, byte *data, int width, int height, unsigned *palette, qboolean alpha)
{
	IDirect3DSurface9 *surf = NULL;

	if (tex && SUCCEEDED (tex->lpVtbl->GetSurfaceLevel (tex, 0, &surf)))
	{
		D3DSURFACE_DESC desc;

		if (SUCCEEDED (surf->lpVtbl->GetDesc (surf, &desc)))
		{
			// if we requested a texture with alpha make sure that it really has alpha
			if (alpha) alpha = R_VerifyTextureAlpha (data, width, height, palette);

			if (desc.Width == width && desc.Height == height)
			{
				// if the size matches upload it directly instead of loading it through D3DXLoadSurfaceFromMemory which may
				// be faster as it can potentially avoid extra allocations and memory copys
				D3DLOCKED_RECT lockrect;

				if (SUCCEEDED (surf->lpVtbl->LockRect (surf, &lockrect, NULL, D3DLOCK_NO_DIRTY_UPDATE)))
				{
					if (palette)
						R_Upload8 ((unsigned *) lockrect.pBits, data, width, height, lockrect.Pitch >> 2, palette, alpha);
					else R_Upload32 ((unsigned *) lockrect.pBits, (unsigned *) data, width, height, lockrect.Pitch >> 2);

					surf->lpVtbl->UnlockRect (surf);
				}
			}
			else
			{
				// general-case slow upload
				RECT srcrect = {0, 0, width, height};

				if (palette)
					data = R_8to32 (data, width, height, palette, alpha);

				// do it this way because it will filter, resample, etc properly for us
				D3DXLoadSurfaceFromMemory (
					surf,
					NULL,
					NULL,
					data,
					D3DFMT_A8R8G8B8,
					width * 4, // * 4 for 32bpp
					NULL,
					&srcrect,
					D3DX_FILTER_LINEAR,
					0);
			}

			// mipmap the texture if necessary (this won't recurse at all if there is only a level-0 miplevel)
			R_GenerateMiplevels (surf, 1);
		}

		// don't leak
		surf->lpVtbl->Release (surf);

		// http://msdn.microsoft.com/en-us/library/windows/desktop/bb172904%28v=vs.85%29.aspx
		// Writing to a non-level-zero surface will not cause the dirty rectangle to be updated. If D3DXLoadSurfaceFromSurface is called and the surface
		// was not already dirty (this is unlikely under normal usage scenarios), the application needs to explicitly call AddDirtyRect on the surface.
		tex->lpVtbl->AddDirtyRect (tex, NULL);

		// preload the managed resource
		tex->lpVtbl->PreLoad (tex);
	}
}


IDirect3DTexture9 *R_CreateTexture (int width, int height, qboolean mipmap, qboolean alpha)
{
	IDirect3DTexture9 *tex = NULL;
	int	scaled_width, scaled_height;

	if (!(d3d_DeviceCaps.TextureCaps & D3DPTEXTURECAPS_NONPOW2CONDITIONAL) && !(d3d_DeviceCaps.TextureCaps & D3DPTEXTURECAPS_POW2))
	{
		// full unconditional non-power-of-two support
		scaled_width = width;
		scaled_height = height;
	}
	else ri->Sys_Error (ERR_FATAL, "(d3d_DeviceCaps.TextureCaps & D3DPTEXTURECAPS_NONPOW2CONDITIONAL) || (d3d_DeviceCaps.TextureCaps & D3DPTEXTURECAPS_POW2)");

	if (d3d_DeviceCaps.TextureCaps & D3DPTEXTURECAPS_SQUAREONLY)
		ri->Sys_Error (ERR_FATAL, "d3d_DeviceCaps.TextureCaps & D3DPTEXTURECAPS_SQUAREONLY");

	// clamp to implementation max
	if (scaled_width > d3d_DeviceCaps.MaxTextureWidth) scaled_width = d3d_DeviceCaps.MaxTextureWidth;
	if (scaled_height > d3d_DeviceCaps.MaxTextureHeight) scaled_height = d3d_DeviceCaps.MaxTextureHeight;

	// and finally try to create it
	if (FAILED (d3d_Device->lpVtbl->CreateTexture (
		d3d_Device,
		scaled_width,
		scaled_height,
		mipmap ? 0 : 1,
		0,
		alpha ? D3DFMT_A8R8G8B8 : D3DFMT_X8R8G8B8,
		D3DPOOL_MANAGED,
		&tex,
		NULL)))
	{
		Sys_Error ("R_CreateTexture : CreateTexture failed");
		return NULL;
	}

	// preload the managed resource
	tex->lpVtbl->PreLoad (tex);

	return tex;
}


void R_LoadSurfaceFromSurface (IDirect3DSurface9 *dst, IDirect3DSurface9 *src)
{
	D3DSURFACE_DESC dstdesc;

	if (SUCCEEDED (dst->lpVtbl->GetDesc (dst, &dstdesc)))
	{
		D3DSURFACE_DESC srcdesc;

		if (SUCCEEDED (src->lpVtbl->GetDesc (src, &srcdesc)))
		{
			if (dstdesc.Width == srcdesc.Width && dstdesc.Height == srcdesc.Height)
			{
				// straight load, no filtering
				D3DXLoadSurfaceFromSurface (dst, NULL, NULL, src, NULL, NULL, D3DX_FILTER_NONE, 0);
			}
			else
			{
				// using gamma correction if we're rescaling
				DWORD filter = D3DX_FILTER_SRGB_IN | D3DX_FILTER_SRGB_OUT;

				// drop back to box filtering if possible so that we'll (hopefully) get a faster load
				if (dstdesc.Width * 2 != srcdesc.Width || dstdesc.Height * 2 != srcdesc.Height)
					filter |= D3DX_FILTER_LINEAR;
				else filter |= D3DX_FILTER_BOX;

				// do it this way because it will filter, resample, etc properly for us
				D3DXLoadSurfaceFromSurface (dst, NULL, NULL, src, NULL, NULL, filter, 0);
			}
		}
	}
}


void R_AlphaEdgeFix (byte *data, int width, int height)
{
	byte *dest = data;

	for (int i = 0; i < height; i++)
	{
		int lastrow = width * 4 * ((i == 0) ? height - 1 : i - 1);
		int thisrow = width * 4 * i;
		int nextrow = width * 4 * ((i == height - 1) ? 0 : i + 1);

		for (int j = 0; j < width; j++, dest += 4)
		{
			// not transparent
			if (dest[3]) continue;

			int lastpix = 4 * ((j == 0) ? width - 1 : j - 1);
			int thispix = 4 * j;
			int nextpix = 4 * ((j == width - 1) ? 0 : j + 1);
			int n = 0, b, c[3] = {0, 0, 0};

			b = lastrow + lastpix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }
			b = thisrow + lastpix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }
			b = nextrow + lastpix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }
			b = lastrow + thispix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }
			b = nextrow + thispix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }
			b = lastrow + nextpix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }
			b = thisrow + nextpix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }
			b = nextrow + nextpix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }

			// average all non-transparent neighbors
			if (n)
			{
				dest[0] = (byte) (c[0] / n);
				dest[1] = (byte) (c[1] / n);
				dest[2] = (byte) (c[2] / n);
			}
		}
	}
}


static const float miptexel[256] = {
	0.000000f, 0.000015f, 0.000062f, 0.000138f, 0.000246f, 0.000384f, 0.000554f, 0.000754f, 0.000984f, 0.001246f, 0.001538f, 0.001861f, 0.002215f, 0.002599f, 0.003014f, 0.003460f,
	0.003937f, 0.004444f, 0.004983f, 0.005552f, 0.006151f, 0.006782f, 0.007443f, 0.008135f, 0.008858f, 0.009612f, 0.010396f, 0.011211f, 0.012057f, 0.012933f, 0.013841f, 0.014779f,
	0.015748f, 0.016747f, 0.017778f, 0.018839f, 0.019931f, 0.021053f, 0.022207f, 0.023391f, 0.024606f, 0.025852f, 0.027128f, 0.028435f, 0.029773f, 0.031142f, 0.032541f, 0.033972f,
	0.035433f, 0.036924f, 0.038447f, 0.040000f, 0.041584f, 0.043199f, 0.044844f, 0.046521f, 0.048228f, 0.049965f, 0.051734f, 0.053533f, 0.055363f, 0.057224f, 0.059116f, 0.061038f,
	0.062991f, 0.064975f, 0.066990f, 0.069035f, 0.071111f, 0.073218f, 0.075356f, 0.077524f, 0.079723f, 0.081953f, 0.084214f, 0.086505f, 0.088827f, 0.091180f, 0.093564f, 0.095978f,
	0.098424f, 0.100900f, 0.103406f, 0.105944f, 0.108512f, 0.111111f, 0.113741f, 0.116401f, 0.119093f, 0.121815f, 0.124567f, 0.127351f, 0.130165f, 0.133010f, 0.135886f, 0.138793f,
	0.141730f, 0.144698f, 0.147697f, 0.150727f, 0.153787f, 0.156878f, 0.160000f, 0.163153f, 0.166336f, 0.169550f, 0.172795f, 0.176071f, 0.179377f, 0.182714f, 0.186082f, 0.189481f,
	0.192910f, 0.196371f, 0.199862f, 0.203383f, 0.206936f, 0.210519f, 0.214133f, 0.217778f, 0.221453f, 0.225160f, 0.228897f, 0.232664f, 0.236463f, 0.240292f, 0.244152f, 0.248043f,
	0.251965f, 0.255917f, 0.259900f, 0.263914f, 0.267959f, 0.272034f, 0.276140f, 0.280277f, 0.284444f, 0.288643f, 0.292872f, 0.297132f, 0.301423f, 0.305744f, 0.310096f, 0.314479f,
	0.318893f, 0.323337f, 0.327812f, 0.332318f, 0.336855f, 0.341423f, 0.346021f, 0.350650f, 0.355310f, 0.360000f, 0.364721f, 0.369473f, 0.374256f, 0.379070f, 0.383914f, 0.388789f,
	0.393695f, 0.398631f, 0.403599f, 0.408597f, 0.413626f, 0.418685f, 0.423775f, 0.428897f, 0.434048f, 0.439231f, 0.444444f, 0.449689f, 0.454964f, 0.460269f, 0.465606f, 0.470973f,
	0.476371f, 0.481799f, 0.487259f, 0.492749f, 0.498270f, 0.503822f, 0.509404f, 0.515017f, 0.520661f, 0.526336f, 0.532042f, 0.537778f, 0.543545f, 0.549343f, 0.555171f, 0.561030f,
	0.566920f, 0.572841f, 0.578793f, 0.584775f, 0.590788f, 0.596832f, 0.602907f, 0.609012f, 0.615148f, 0.621315f, 0.627513f, 0.633741f, 0.640000f, 0.646290f, 0.652611f, 0.658962f,
	0.665344f, 0.671757f, 0.678201f, 0.684675f, 0.691180f, 0.697716f, 0.704283f, 0.710880f, 0.717509f, 0.724168f, 0.730857f, 0.737578f, 0.744329f, 0.751111f, 0.757924f, 0.764767f,
	0.771642f, 0.778547f, 0.785483f, 0.792449f, 0.799446f, 0.806474f, 0.813533f, 0.820623f, 0.827743f, 0.834894f, 0.842076f, 0.849289f, 0.856532f, 0.863806f, 0.871111f, 0.878447f,
	0.885813f, 0.893210f, 0.900638f, 0.908097f, 0.915586f, 0.923106f, 0.930657f, 0.938239f, 0.945852f, 0.953495f, 0.961169f, 0.968874f, 0.976609f, 0.984375f, 0.992172f, 1.000000f
};


int AverageMip (int _1, int _2, int _3, int _4)
{
	return (_1 + _2 + _3 + _4) >> 2;
}


int AverageMipGC (int _1, int _2, int _3, int _4)
{
	return (int) (255.0f * sqrt ((miptexel[_1] + miptexel[_2] + miptexel[_3] + miptexel[_4]) / 4.0f));
}


unsigned *R_ResampleTexture2 (unsigned *in, int inwidth, int inheight, int outwidth, int outheight)
{
	// can this ever happen???
	if (outwidth == inwidth && outheight == inheight) return in;

	// allocating the out buffer before the pool mark so that we can return it
	unsigned *out = (unsigned *) ri->Scratch_AllocSize (outwidth * outheight * 4);

	// and now get the mark because all allocations after this will be released
	int mark = ri->Scratch_LowMark ();

	unsigned *p1 = (unsigned *) ri->Scratch_AllocSize (outwidth * 4);
	unsigned *p2 = (unsigned *) ri->Scratch_AllocSize (outwidth * 4);

	unsigned fracstep = inwidth * 0x10000 / outwidth;
	unsigned frac = fracstep >> 2;

	for (int i = 0; i < outwidth; i++)
	{
		p1[i] = 4 * (frac >> 16);
		frac += fracstep;
	}

	frac = 3 * (fracstep >> 2);

	for (int i = 0; i < outwidth; i++)
	{
		p2[i] = 4 * (frac >> 16);
		frac += fracstep;
	}

	for (int i = 0; i < outheight; i++)
	{
		unsigned *outrow = out + (i * outwidth);
		unsigned *inrow0 = in + inwidth * (int) (((i + 0.25f) * inheight) / outheight);
		unsigned *inrow1 = in + inwidth * (int) (((i + 0.75f) * inheight) / outheight);

		for (int j = 0; j < outwidth; j++)
		{
			byte *pix1 = (byte *) inrow0 + p1[j];
			byte *pix2 = (byte *) inrow0 + p2[j];
			byte *pix3 = (byte *) inrow1 + p1[j];
			byte *pix4 = (byte *) inrow1 + p2[j];

			// don't gamma correct the alpha channel
			((byte *) &outrow[j])[0] = AverageMipGC (pix1[0], pix2[0], pix3[0], pix4[0]);
			((byte *) &outrow[j])[1] = AverageMipGC (pix1[1], pix2[1], pix3[1], pix4[1]);
			((byte *) &outrow[j])[2] = AverageMipGC (pix1[2], pix2[2], pix3[2], pix4[2]);
			((byte *) &outrow[j])[3] = AverageMip (pix1[3], pix2[3], pix3[3], pix4[3]);
		}
	}

	ri->Scratch_FreeToLowMark (mark);

	return out;
}


unsigned *R_ResampleTexture (unsigned *in, int inwidth, int inheight)
{
	// round down to meet np2 specification
	return R_ResampleTexture2 (in, inwidth, inheight, (inwidth > 1) ? (inwidth >> 1) : 1, (inheight > 1) ? (inheight >> 1) : 1);
}


unsigned *R_MipMap (unsigned *data, int width, int height)
{
	// because each SRD must have it's own data we can't mipmap in-place otherwise we'll corrupt the previous miplevel
	unsigned *trans = (unsigned *) ri->Scratch_AllocSize ((width >> 1) * (height >> 1) * 4);
	byte *in = (byte *) data;
	byte *out = (byte *) trans;

	// do this after otherwise it will interfere with the allocation size above
	width <<= 2;
	height >>= 1;

	for (int i = 0; i < height; i++, in += width)
	{
		for (int j = 0; j < width; j += 8, out += 4, in += 8)
		{
			// don't gamma correct the alpha channel
			out[0] = AverageMipGC (in[0], in[4], in[width + 0], in[width + 4]);
			out[1] = AverageMipGC (in[1], in[5], in[width + 1], in[width + 5]);
			out[2] = AverageMipGC (in[2], in[6], in[width + 2], in[width + 6]);
			out[3] = AverageMip (in[3], in[7], in[width + 3], in[width + 7]);
		}
	}

	return trans;
}


#define TEX_MIPMAP 1

void R_GenerateTextureMipmaps (unsigned *trans, int scaled_width, int scaled_height, int miplevel, int flags)
{
	// lower-bound
	if (scaled_width < 1) scaled_width = 1;
	if (scaled_height < 1) scaled_height = 1;

	// upload this level
	//glTexImage2D (GL_TEXTURE_2D, miplevel, GL_RGBA8, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, trans);

	// upload the next level (if mipmapped)
	if ((scaled_width > 1 || scaled_height > 1) && (flags & TEX_MIPMAP))
	{
		// choose the appropriate filter
		if ((scaled_width & 1) || (scaled_height & 1))
			trans = R_ResampleTexture (trans, scaled_width, scaled_height);
		else trans = R_MipMap (trans, scaled_width, scaled_height);

		R_GenerateTextureMipmaps (trans, scaled_width >> 1, scaled_height >> 1, miplevel + 1, flags);
	}
}



void R_FillTexture666 (IDirect3DTexture9 *tex, byte *data, int width, int height, unsigned *palette, qboolean alpha)
{
	// because we're going to be doing lots of allocations
	int mark = ri->Scratch_LowMark ();

	if (palette)
	{
		// upsample to 32-bit
		// because this doesn't advance the lowmark it's not safe to do a further allocation until after the data has been consumed
		unsigned *trans = (unsigned *) ri->Scratch_AllocSize (width * height * 4);
		unsigned *dst = trans;

		// send through R_Upload8 so that we get consistent alpha handling on both paths
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				// 255 is alpha, even in custom palettes
				if (alpha && data[x] == 255)
					dst[x] = 0;
				else dst[x] = palette[data[x]];
			}

			data += width;
			dst += width;
		}

		// copy back and NULL-out palette so any further attempt to access it is an error
		data = (byte *) trans;
	}
}


