/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_light.c

#include "gl_local.h"


#define	DLIGHT_CUTOFF	64

// incremented each bmodel pass, only surfs marked get dynamic lights
int r_dlightframecount = 1;

// scaled lightstyles
float d_lightstyles[MAX_LIGHTSTYLES];
qboolean ls_dirty = false;

void R_UpdateLightStyles (refdef_t *rd)
{
	// a RDF_NOWORLDMODEL refdef won't have styles
	if (rd->rdflags & RDF_NOWORLDMODEL) return;

	// scale the lightstyles for gl_modulate and overbrighting range
	for (int i = 0; i < MAX_LIGHTSTYLES; i++)
	{
		// i had "style *= style" here to boost visibility of lightstyles, but it led to some excessive over-
		// brightening in the warehouse (after the lights come on with the first battery) so i backed it out.
		// the rationalization was that this was a change with the same philosophy as "double the size of eyes
		// because they're harder to see in GL" so i was OK in principle with it up to that point.
		float style = (gl_dynamic->value > 0.0f) ? ((rd->lightstyles[i] - 1.0f) * gl_dynamic->value) + 1.0f : 1.0f;

		// scale for overbrighting
		style *= gl_modulate->value;

		// check for a change, store it back, and flag the change
		if (style != d_lightstyles[i])
		{
			d_lightstyles[i] = style;
			ls_dirty = true;
		}
	}

	// now update any styles which were dirty
	// (tradeoff: a single batch upload even including styles which were not dirty versus multiple tiny uploads)
	if (ls_dirty)
	{
		R_SetVertexShaderConstant4fv (VSC_LIGHTSTYLES, d_lightstyles, MAX_PACKED_LIGHTSTYLES);
		ls_dirty = false;
	}
}


void R_DirtyLightStyles (void)
{
	// force all styles to dirty for the next update
	for (int i = 0; i < MAX_LIGHTSTYLES; i++)
		d_lightstyles[i] = gl_modulate->value;

	ls_dirty = true;
}


/*
=============================================================================

DYNAMIC LIGHTS

=============================================================================
*/

void R_DrawDLightChains (dlight_t *dl);
void R_ClearTextureChains (void);

/*
=============
R_MarkLights
=============
*/
static void R_MarkLights (dlight_t *light, mnode_t *node)
{
	if (node->contents != -1) return;

	cplane_t *splitplane = node->plane;
	float dist = Vector3Dot (light->transformed, splitplane->normal) - splitplane->dist;

	// don't cut off the dlight with GPU lightmaps
	if (dist > light->intensity)
	{
		R_MarkLights (light, node->children[0]);
		return;
	}

	// don't cut off the dlight with GPU lightmaps
	if (dist < -light->intensity)
	{
		R_MarkLights (light, node->children[1]);
		return;
	}

	// mark the polygons
	float maxdist = light->intensity * light->intensity;
	msurface_t *surf = node->surfaces;

	for (int i = 0; i < node->numsurfaces; i++, surf++)
	{
		// no dlights on these surfaces
		if (surf->texinfo->flags & (SURF_SKY | SURF_TRANS33 | SURF_TRANS66 | SURF_WARP)) continue;

		// the surface was not drawn in the main render so don't bother adding lights to it
		// otherwise backfacing surfs or surfs not in the PVS may be added to these passes by mistake
		if (surf->dlightframe != r_dlightframecount) continue;

		// with scissor optimizations it's now faster to save some CPU but burn the extra GPU on surfs that may not be lit
		// this surface is now to be lit so add it to the appropriate chain
		image_t *image = R_TextureAnimation (surf->texinfo, surf->animframe);

		surf->texturechain = image->texturechain;
		image->texturechain = surf;

		// mark the light as active so that we'll know to draw it
		light->active = true;
	}

	R_MarkLights (light, node->children[0]);
	R_MarkLights (light, node->children[1]);
}


qboolean R_ShouldEntityHaveDlights (entity_t *e)
{
	// no lights
	if (!(gl_dynamic->value > 0.0f) || r_fullbright->value) return false;
	if (e->flags & (RF_TRANSLUCENT | RF_FULLBRIGHT)) return false;
	if (e->flags & (RF_SHELL_RED | RF_SHELL_GREEN | RF_SHELL_BLUE | RF_SHELL_DOUBLE | RF_SHELL_HALF_DAM)) return false;
	if (!(e->flags & RF_DYNAMICLIGHT)) return false;

	// the entity has lights now
	return true;
}

/*
=============
R_PushDlights
=============
*/
void R_PushDlights (entity_t *e, struct mnode_s *headnode)
{
	if (!R_ShouldEntityHaveDlights (e) || e->model->firstnode < 0)
	{
		// bump the dlight frame so that the next pass won't add surfs it shouldn't
		r_dlightframecount++;
		return;
	}

	for (int i = 0; i < r_newrefdef.num_dlights; i++)
	{
		dlight_t *dl = &r_newrefdef.dlights[i];

		// dlight is initially inactive
		dl->active = false;

		if (dl->intensity < DLIGHT_CUTOFF) continue;

		// move the dlight back to entity space
		R_InverseTransform (&e->localmatrix, dl->transformed, dl->origin);

		// clear down the texture chains
		R_ClearTextureChains ();

		// mark surfaces for it
		R_MarkLights (dl, headnode);

		// nothing was hit
		if (!dl->active) continue;

		// setup this light
		if (!R_BeginEntityLighting (e, dl)) continue;

		// and draw marked surfaces
		R_DrawDLightChains (dl);
	}

	R_EndEntityLighting (e);

	// bump the dlight frame so that the next pass won't add surfs it shouldn't
	r_dlightframecount++;
}


int R_ScissorLight (RECT *sr, float *lightpos, float r)
{
	/*
	http://www.gamasutra.com/view/feature/131351/the_mechanics_of_robust_stencil_.php?page=6
	ugh - lengyel stuff.  this is going to be full of random equations he's pulled out of his arse that
	are significantly more complex than they need to be, and no indication of how they relate to a
	real-world usage.  the entire thing can probably be written in 2 lines of simple and clear C.
	i got this from a post on the opengl.org forums that the author claims is a working implementation,
	and bashed it into shape for Quake and D3D.
	*/
	int rect[4] = {0, 0, vid.width, vid.height};
	float d, l[3];

	// move the light to eye space
	R_Transform (&r_view_matrix, l, lightpos);

	float r2 = r * r;		// squared radius
	float l2[3] = {l[0] * l[0], l[1] * l[1], l[2] * l[2]};

	// derive from the projection matrix
	float e1 = r_projection_matrix.m4x4[0][0];
	float e2 = r_projection_matrix.m4x4[1][1];

	d = r2 * l2[0] - (l2[0] + l2[2]) * (r2 - l2[2]);

	if (d >= 0)
	{
		d = sqrtf (d);

		float nx1 = (r * l[0] + d) / (l2[0] + l2[2]);
		float nx2 = (r * l[0] - d) / (l2[0] + l2[2]);

		float nz1 = (r - nx1 * l[0]) / l[2];
		float nz2 = (r - nx2 * l[0]) / l[2];

		float pz1 = (l2[0] + l2[2] - r2) / (l[2] - (nz1 / nx1) * l[0]);
		float pz2 = (l2[0] + l2[2] - r2) / (l[2] - (nz2 / nx2) * l[0]);

		if (pz1 >= 0 && pz2 >= 0)
			return 0; // discard light

		if (pz1 < 0)
		{
			float fx = nz1 * e1 / nx1;
			int ix = (int) ((fx + 1.0f) * vid.width * 0.5f);
			float px = -pz1 * nz1 / nx1;

			if (px < l[0])
				rect[0] = max (rect[0], ix);
			else rect[2] = min (rect[2], ix);
		}

		if (pz2 < 0)
		{
			float fx = nz2 * e1 / nx2;
			int ix = (int) ((fx + 1.0f) * vid.width * 0.5f);
			float px = -pz2 * nz2 / nx2;

			if (px < l[0])
				rect[0] = max (rect[0], ix);
			else rect[2] = min (rect[2], ix);
		}

		if (rect[2] <= rect[0])
			return 0; // discard light
	}

	d = r2 * l2[1] - (l2[1] + l2[2]) * (r2 - l2[2]);

	if (d >= 0)
	{
		d = sqrtf (d);

		float ny1 = (r * l[1] + d) / (l2[1] + l2[2]);
		float ny2 = (r * l[1] - d) / (l2[1] + l2[2]);

		float nz1 = (r - ny1 * l[1]) / l[2];
		float nz2 = (r - ny2 * l[1]) / l[2];

		float pz1 = (l2[1] + l2[2] - r2) / (l[2] - (nz1 / ny1) * l[1]);
		float pz2 = (l2[1] + l2[2] - r2) / (l[2] - (nz2 / ny2) * l[1]);

		if (pz1 >= 0 && pz2 >= 0)
			return 0; // discard light

		if (pz1 < 0)
		{
			float fy = nz1 * e2 / ny1;
			int iy = (int) ((fy + 1.0f) * vid.height * 0.5f);
			float py = -pz1 * nz1 / ny1;

			if (py < l[1])
				rect[1] = max (rect[1], iy);
			else rect[3] = min (rect[3], iy);
		}

		if (pz2 < 0)
		{
			float fy = nz2 * e2 / ny2;
			int iy = (int) ((fy + 1.0f) * vid.height * 0.5f);
			float py = -pz2 * nz2 / ny2;

			if (py < l[1])
				rect[1] = max (rect[1], iy);
			else rect[3] = min (rect[3], iy);
		}

		if (rect[3] <= rect[1])
			return 0; // discard light
	}

	int n = (rect[2] - rect[0]) * (rect[3] - rect[1]);

	if (n)
	{
		// assign the rect
		sr->left = rect[0];
		sr->right = rect[2];

		// fixup opengl shitty coordinate system because we're drawing on a screen, not on graph paper
		sr->top = vid.height - rect[3];
		sr->bottom = vid.height - rect[1];
	}

	return n;
}


BOOL R_BeginEntityLighting (entity_t *e, dlight_t *dl)
{
	// clear dynamic lighting flags
	e->flags &= ~(RF_ADDITIVEBLEND | RF_SUBTRACTBLEND);

	// test the scissor rect
	// if the light was fully scissored out, don't do it
	if (dl->rect.left >= dl->rect.right) return FALSE;
	if (dl->rect.top >= dl->rect.bottom) return FALSE;

	// set up the scissor rect
	R_SetScissor (TRUE, &dl->rect);

	// go to the appropriate blending mode
	if (dl->negative)
		e->flags |= RF_SUBTRACTBLEND;
	else e->flags |= RF_ADDITIVEBLEND;

	// and push the new state bits
	R_SetEntityState (e);

	// scale for overbright, modulate and dynamic
	R_SetPixelShaderConstant4f (
		PSC_LIGHTCOLOR,
		dl->color[0] * gl_modulate->value * gl_dynamic->value,
		dl->color[1] * gl_modulate->value * gl_dynamic->value,
		dl->color[2] * gl_modulate->value * gl_dynamic->value,
		0.0f
	);

	// and send the rest of them
	R_SetPixelShaderConstant1f (PSC_LIGHTRADIUS, dl->intensity);
	R_SetVertexShaderConstant4fv (VSC_TRANSFORMEDDL, dl->transformed, 1);

	// the light is good now
	return TRUE;
}


void R_EndEntityLighting (entity_t *e)
{
	// clear dynamic lighting flags
	e->flags &= ~(RF_ADDITIVEBLEND | RF_SUBTRACTBLEND);

	// no more scissoring
	R_SetScissor (FALSE, NULL);
}


void R_RunDLights (refdef_t *rd)
{
	for (int i = 0; i < rd->num_dlights; i++)
	{
		dlight_t *dl = &rd->dlights[i];

		// if the scissor determines that the light is offscreen we set an inverted rect so that the drawing code can detect it
		// (this is something we want to detect in the general case anyway)
		if (!R_ScissorLight (&dl->rect, dl->origin, dl->intensity))
			SetRect (&dl->rect, 1, 1, 0, 0);
	}
}


/*
=============================================================================

LIGHT SAMPLING

=============================================================================
*/

vec3_t			pointcolor;
cplane_t		*lightplane;		// used as shadow plane
vec3_t			lightspot;

int R_RecursiveLightPoint (mnode_t *node, vec3_t start, vec3_t end)
{
	float		front, back, frac;
	int			side;
	cplane_t	*plane;
	vec3_t		mid;
	msurface_t	*surf;
	int			s, t, ds, dt;
	int			i;
	mtexinfo_t	*tex;
	byte		*lightmap;
	int			maps;
	int			r;

	// didn't hit anything
	if (node->contents != -1) return -1;

	// calculate mid point
	plane = node->plane;
	front = Vector3Dot (start, plane->normal) - plane->dist;
	back = Vector3Dot (end, plane->normal) - plane->dist;
	side = front < 0;

	if ((back < 0) == side)
		return R_RecursiveLightPoint (node->children[side], start, end);

	frac = front / (front - back);
	mid[0] = start[0] + (end[0] - start[0]) * frac;
	mid[1] = start[1] + (end[1] - start[1]) * frac;
	mid[2] = start[2] + (end[2] - start[2]) * frac;

	// go down front side
	r = R_RecursiveLightPoint (node->children[side], start, mid);

	if (r >= 0) return r;					// hit something
	if ((back < 0) == side) return -1;		// didn't hit anuthing

	// check for impact on this node
	VectorCopy (mid, lightspot);
	lightplane = plane;

	surf = node->surfaces;

	for (i = 0; i < node->numsurfaces; i++, surf++)
	{
		// no lightmaps
		if (surf->texinfo->flags & (SURF_SKY | SURF_TRANS33 | SURF_TRANS66 | SURF_WARP)) continue;

		tex = surf->texinfo;

		s = Vector3Dot (mid, tex->vecs[0]) + tex->vecs[0][3];
		t = Vector3Dot (mid, tex->vecs[1]) + tex->vecs[1][3];;

		if (s < surf->texturemins[0] || t < surf->texturemins[1]) continue;

		ds = s - surf->texturemins[0];
		dt = t - surf->texturemins[1];

		if (ds > surf->extents[0] || dt > surf->extents[1]) continue;

		// clear the colour before checking surf->samples (this was a bug from Q1 where the return value was the light colour; in Q2
		// it's still a 0 light colour so it's a valid hit, but returning 0 won't clear the color so we need to do so ourselves....)
		VectorCopy (vec3_origin, pointcolor);

		if (!surf->samples) return 0;

		ds >>= 4;
		dt >>= 4;

		lightmap = surf->samples;

		if (lightmap)
		{
			lightmap += 3 * (dt * ((surf->extents[0] >> 4) + 1) + ds);

			for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
			{
				float scale = d_lightstyles[surf->styles[maps]] * (1.0 / 255.0f);

				// doubled for overbrighting range
				pointcolor[0] += lightmap[0] * scale;
				pointcolor[1] += lightmap[1] * scale;
				pointcolor[2] += lightmap[2] * scale;

				lightmap += 3 * ((surf->extents[0] >> 4) + 1) * ((surf->extents[1] >> 4) + 1);
			}
		}

		return 1;
	}

	// go down back side
	return R_RecursiveLightPoint (node->children[!side], mid, end);
}

/*
===============
R_LightPoint
===============
*/
void R_LightPoint (vec3_t p, vec3_t color)
{
	vec3_t end, hit, lightspot2;

	if (!r_worldmodel->lightdata)
	{
		Vector3Set (color, 1.0f, 1.0f, 1.0f);
		return;
	}

	// don't go outside of the world's bounds
	Vector3Set (end, p[0], p[1], r_worldmodel->mins[2] - 10.0f);

	if ((R_RecursiveLightPoint (r_worldmodel->nodes, p, end)) < 0)
	{
		// objects outside the world (such as Strogg Viper flybys on base1) should be lit if they hit nothing
		// the new end point (based on mins of r_worldmodel) ensures that we'll always hit if inside the world so this is OK
		// don't make them too bright or they'll look like shit/
		Vector3Set (color, 0.5f, 0.5f, 0.5f);
		Vector3Copy (hit, end);
	}
	else
	{
		Vector3Copy (color, pointcolor);
		Vector3Copy (hit, lightspot);
	}

	// find bmodels under the lightpoint - move the point to bmodel space, trace down, then check; if r < 0
	// it didn't find a bmodel, otherwise it did (a bmodel under a valid world hit will hit here too)
	// fixme: is it possible for a bmodel to not be in the PVS but yet be a valid candidate for this???
	for (int i = 0; i < r_newrefdef.num_entities; i++)
	{
		entity_t *e = &r_newrefdef.entities[i];
		float estart[3], eend[3];

		// beam and translucent ents don't light
		if (e->flags & RF_BEAM) continue;
		if (e->flags & RF_TRANSLUCENT) continue;

		// NULL models don't light
		if (!e->model) continue;

		// only bmodel entities give light
		if (e->model->type != mod_brush) continue;

		// this happens on boss1
		if (e->model->firstnode < 0) continue;

		// move start and end points into the entity's frame of reference
		R_InverseTransform (&e->localmatrix, estart, p);
		R_InverseTransform (&e->localmatrix, eend, end);

		// and run the recursive light point on it too
		if (!(R_RecursiveLightPoint (e->model->nodes + e->model->firstnode, estart, eend) < 0))
		{
			// a bmodel under a valid world hit will hit here too so take the highest lightspot on all hits
			// move lightspot back to world space
			R_Transform (&e->localmatrix, lightspot2, lightspot);

			if (lightspot2[2] > hit[2])
			{
				// found a bmodel so copy it over
				Vector3Copy (color, pointcolor);
				Vector3Copy (hit, lightspot2);
			}
		}
	}

	// the final hit point is the valid lightspot
	Vector3Copy (lightspot, hit);
}


void R_DynamicLightPoint (vec3_t p, vec3_t color)
{
	// the main passes for alias models don't add dynamics because they use realtime lighting instead;
	// passes for null models and setting r_lightlevel do
	if (gl_dynamic->value > 0.0f)
	{
		for (int lnum = 0; lnum < r_newrefdef.num_dlights; lnum++)
		{
			dlight_t *dl = &r_newrefdef.dlights[lnum];
			float dist[3], add;

			// this was "currententity->origin" instead of "p", meaning that it never worked for R_SetLightLevel
			VectorSubtract (p, dl->origin, dist);

			add = dl->intensity - VectorLength (dist);
			add *= (1.0 / 256);
			add *= gl_modulate->value;
			add *= gl_dynamic->value;

			if (add > 0)
				VectorMA (color, add, dl->color, color);
		}
	}
}


/*
====================
R_SetLightLevel

====================
*/
void R_SetLightLevel (void)
{
	vec3_t	shadelight;

	if (r_newrefdef.rdflags & RDF_NOWORLDMODEL)
		return;

	// save off light value for server to look at (BIG HACK!)
	// this never worked right because R_LightPoint used currententity->origin instead of p for stuff
	// but currententity would not have been set right when this was called - it does work right now.
	// (that's probably why it was also added to R_DrawAliasModel)
	R_LightPoint (r_newrefdef.vieworigin, shadelight);
	R_DynamicLightPoint (r_newrefdef.vieworigin, shadelight);

	// pick the greatest component, which should be the same
	// as the mono value returned by software
	float best = -9999999.0f;

	if (shadelight[0] > best) best = shadelight[0];
	if (shadelight[1] > best) best = shadelight[1];
	if (shadelight[2] > best) best = shadelight[2];

	r_lightlevel->value = 150.0f * best;
}


