/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_light.c

#include "gl_local.h"

#define	MAX_LIGHTMAPS	256
#define	LIGHTMAP_SIZE	256

static image_t d3d_Lightmaps[MAX_LIGHTMAPS][MAXLIGHTMAPS];


typedef struct gllightmapstate_s
{
	int	current_lightmap_texture;
	int	allocated[LIGHTMAP_SIZE];
} gllightmapstate_t;


static gllightmapstate_t gl_lms;


#if 1
/*
===============
R_BuildLightMap

Combine and scale multiple lightmaps into the floating format in blocklights

this can be done with 3 textures, loading each of r, g and b into it's own texture then using each of the
colour channels in a texture for the styles.  then draw it single-passed but with 3 texture reads.  use a
texture array to allow us to bind just once.  might be faster, might not - a tradeoff to evaluate...
===============
*/
void R_BuildLightMap (msurface_t *surf, image_t *lm)
{
	if (surf->texinfo->flags & (SURF_SKY | SURF_TRANS33 | SURF_TRANS66 | SURF_WARP))
		ri->Sys_Error (ERR_DROP, "R_BuildLightMap called for non-lit surface");

	int smax = (surf->extents[0] >> 4) + 1;
	int tmax = (surf->extents[1] >> 4) + 1;
	byte *lightmap = surf->samples;

	for (int maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
	{
		// create the texture if needed
		if (!lm[maps].texture) lm[maps].texture = R_CreateTexture (LIGHTMAP_SIZE, LIGHTMAP_SIZE, false, false);

		IDirect3DTexture9 *tex = lm[maps].texture;
		D3DLOCKED_RECT lockrect;
		RECT surfrect = {surf->light_s, surf->light_t, surf->light_s + smax, surf->light_t + tmax};

		// this is only done at level load nowadays; using D3DLOCK_NO_DIRTY_UPDATE so we won't incur the cost of multiple
		// tiny uploads during building (we can happily incur it in bulk post-upload via AddDirtyRect)
		if (SUCCEEDED (tex->lpVtbl->LockRect (tex, 0, &lockrect, &surfrect, D3DLOCK_NO_DIRTY_UPDATE)))
		{
			unsigned *rgba = (unsigned *) lockrect.pBits;
			int stride = lockrect.Pitch >> 2;

			for (int i = 0; i < tmax; i++, rgba += stride)
			{
				for (int j = 0; j < smax; j++)
				{
					// flip to BGR (do this at map load time???)
					if (surf->samples)
					{
						((byte *) &rgba[j])[2] = *lightmap++;
						((byte *) &rgba[j])[1] = *lightmap++;
						((byte *) &rgba[j])[0] = *lightmap++;
					}
					else
					{
						((byte *) &rgba[j])[0] = 128;
						((byte *) &rgba[j])[1] = 128;
						((byte *) &rgba[j])[2] = 128;
					}
				}
			}

			tex->lpVtbl->UnlockRect (tex, 0);
		}
	}
}
#else
unsigned short R_R8G8B8_to_R5G6B5 (int r, int g, int b)
{
	// load lightmaps as 16-bit colour because they can take the quality reduction more easily
	return ((r >> 3) << 11) | ((g >> 2) << 5) | (b >> 3);
}


/*
===============
R_BuildLightMap

Combine and scale multiple lightmaps into the floating format in blocklights

this can be done with 3 textures, loading each of r, g and b into it's own texture then using each of the
colour channels in a texture for the styles.  then draw it single-passed but with 3 texture reads.  use a
texture array to allow us to bind just once.  might be faster, might not - a tradeoff to evaluate...
===============
*/
void R_BuildLightMap (msurface_t *surf, image_t *lm)
{
	if (surf->texinfo->flags & (SURF_SKY | SURF_TRANS33 | SURF_TRANS66 | SURF_WARP))
		ri->Sys_Error (ERR_DROP, "R_BuildLightMap called for non-lit surface");

	int smax = (surf->extents[0] >> 4) + 1;
	int tmax = (surf->extents[1] >> 4) + 1;
	byte *lightmap = surf->samples;

	for (int maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
	{
		// create the texture if needed
		if (!lm[maps].texture)
		{
			d3d_Device->lpVtbl->CreateTexture (
				d3d_Device,
				LIGHTMAP_SIZE,
				LIGHTMAP_SIZE,
				1,
				0,
				D3DFMT_R5G6B5,
				D3DPOOL_MANAGED,
				&lm[maps].texture,
				NULL
			);
		}

		IDirect3DTexture9 *tex = lm[maps].texture;
		D3DLOCKED_RECT lockrect;
		RECT surfrect = {surf->light_s, surf->light_t, surf->light_s + smax, surf->light_t + tmax};

		// this is only done at level load nowadays; using D3DLOCK_NO_DIRTY_UPDATE so we won't incur the cost of multiple
		// tiny uploads during building (we can happily incur it in bulk post-upload via AddDirtyRect)
		if (SUCCEEDED (tex->lpVtbl->LockRect (tex, 0, &lockrect, &surfrect, D3DLOCK_NO_DIRTY_UPDATE)))
		{
			unsigned short *rgba = (unsigned short *) lockrect.pBits;
			int stride = lockrect.Pitch >> 1;

			for (int i = 0; i < tmax; i++, rgba += stride)
			{
				for (int j = 0; j < smax; j++)
				{
					// compress lightmaps to 16-bit colour so that we don't use too much memory
					if (surf->samples)
					{
						rgba[j] = R_R8G8B8_to_R5G6B5 (lightmap[0], lightmap[1], lightmap[2]);
						lightmap += 3;
					}
					else rgba[j] = R_R8G8B8_to_R5G6B5 (128, 128, 128);
				}
			}

			tex->lpVtbl->UnlockRect (tex, 0);
		}
	}
}
#endif


static void LM_InitBlock (void)
{
	memset (gl_lms.allocated, 0, sizeof (gl_lms.allocated));
}


static void LM_UploadBlock (void)
{
	image_t *lm = d3d_Lightmaps[gl_lms.current_lightmap_texture];

	for (int i = 0; i < 4; i++)
	{
		if (lm[i].texture)
		{
			// perform the final upload via AddDirtyRect
			lm[i].texture->lpVtbl->AddDirtyRect (lm[i].texture, NULL);
			lm[i].texture->lpVtbl->PreLoad (lm[i].texture);

			// and register it
			lm[i].registration_sequence = r_registration_sequence;
		}
	}

	if (++gl_lms.current_lightmap_texture >= MAX_LIGHTMAPS)
		ri->Sys_Error (ERR_DROP, "LM_UploadBlock () - MAX_LIGHTMAPS exceeded\n");
}


// returns a texture number and the position inside it
static qboolean LM_AllocBlock (int w, int h, int *x, int *y)
{
	int best = LIGHTMAP_SIZE;

	for (int i = 0; i < LIGHTMAP_SIZE - w; i++)
	{
		int j, best2 = 0;

		for (j = 0; j < w; j++)
		{
			if (gl_lms.allocated[i + j] >= best)
				break;

			if (gl_lms.allocated[i + j] > best2)
				best2 = gl_lms.allocated[i + j];
		}

		if (j == w)
		{
			// this is a valid spot
			*x = i;
			*y = best = best2;
		}
	}

	if (best + h > LIGHTMAP_SIZE)
	{
		LM_UploadBlock ();
		LM_InitBlock ();
		return LM_AllocBlock (w, h, x, y);
	}

	for (int i = 0; i < w; i++)
		gl_lms.allocated[*x + i] = best + h;

	return true;
}


/*
========================
R_CreateSurfaceLightmap
========================
*/
void R_CreateSurfaceLightmap (msurface_t *surf)
{
	if (surf->texinfo->flags & (SURF_SKY | SURF_TRANS33 | SURF_TRANS66 | SURF_WARP)) return;

	int smax = (surf->extents[0] >> 4) + 1;
	int tmax = (surf->extents[1] >> 4) + 1;

	LM_AllocBlock (smax, tmax, &surf->light_s, &surf->light_t);

	surf->lightmaptexturenum = gl_lms.current_lightmap_texture;
	R_BuildLightMap (surf, d3d_Lightmaps[surf->lightmaptexturenum]);
}


/*
==================
R_BeginBuildingLightmaps

==================
*/
void R_BeginBuildingLightmaps (model_t *m)
{
	// nothing allocated yet
	gl_lms.current_lightmap_texture = 0;

	// clear all allocations
	LM_InitBlock ();
}


/*
=======================
R_EndBuildingLightmaps
=======================
*/
void R_EndBuildingLightmaps (void)
{
	// upload the last block that was filled
	LM_UploadBlock ();

	// free all lightmaps that weren't used this registration sequence
	R_ReleaseLightmaps (r_registration_sequence);

	// dirty all the lightstyles so we'll get a full update in the first frame
	R_DirtyLightStyles ();
}


void R_ClearLightmapChains (void)
{
	for (int i = 0; i < gl_lms.current_lightmap_texture; i++)
	{
		d3d_Lightmaps[i][0].texturechain = NULL;
		d3d_Lightmaps[i][1].texturechain = NULL;
		d3d_Lightmaps[i][2].texturechain = NULL;
		d3d_Lightmaps[i][3].texturechain = NULL;
	}
}


void R_ReleaseLightmaps (int rs)
{
	for (int i = 0; i < MAX_LIGHTMAPS; i++)
	{
		// release any textures that were not used in this sequence
		if (d3d_Lightmaps[i][0].registration_sequence != rs) R_ReleaseTexture (&d3d_Lightmaps[i][0]);
		if (d3d_Lightmaps[i][1].registration_sequence != rs) R_ReleaseTexture (&d3d_Lightmaps[i][1]);
		if (d3d_Lightmaps[i][2].registration_sequence != rs) R_ReleaseTexture (&d3d_Lightmaps[i][2]);
		if (d3d_Lightmaps[i][3].registration_sequence != rs) R_ReleaseTexture (&d3d_Lightmaps[i][3]);
	}
}


void R_BuildLightmapChain (msurface_t *surf, int pass)
{
	image_t *lm = &d3d_Lightmaps[surf->lightmaptexturenum][pass];

	surf->lightmapchain = lm->texturechain;
	lm->texturechain = surf;
}


void R_DrawLightmappedChain (int pass)
{
	for (int i = 0; i < gl_lms.current_lightmap_texture; i++)
	{
		image_t *lm = &d3d_Lightmaps[i][pass];
		surfbatch_t batch;

		if (!lm->texture) continue;
		if (!lm->texturechain) continue;

		if (r_fullbright->value)
			R_SetTexture (TS_LIGHT, r_greytexture);
		else R_SetTexture (TS_LIGHT, lm);

		R_BeginBatchingSurfaces (&batch);

		for (msurface_t *surf = lm->texturechain; surf; surf = surf->lightmapchain)
			R_BatchSurface (&batch, surf, NULL);

		R_EndBatchingSurfaces (&batch);
		lm->texturechain = NULL;
	}
}


void R_SetLightmapTexCoords (float *lm, msurface_t *surf, float *vec)
{
	// lightmap texture coordinates
	mtexinfo_t *ti = surf->texinfo;

	lm[0] = ((Vector3Dot (vec, ti->vecs[0]) + ti->vecs[0][3]) - surf->texturemins[0] + (surf->light_s << 4) + 8.0f) / (LIGHTMAP_SIZE << 4);
	lm[1] = ((Vector3Dot (vec, ti->vecs[1]) + ti->vecs[1][3]) - surf->texturemins[1] + (surf->light_t << 4) + 8.0f) / (LIGHTMAP_SIZE << 4);
}


