/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_sky.c -- sky and water polygons

#include "gl_local.h"
#include "resource.h"

char	skyname[MAX_QPATH];
float	skyrotate;
vec3_t	skyaxis;
image_t	*sky_images[6];

IDirect3DCubeTexture9 *d3d_SkyCubemap = NULL;

IDirect3DVertexShader9 *d3d_SurfSkyVS = NULL;
IDirect3DPixelShader9 *d3d_SurfSkyPS = NULL;

void Sky_OnCreate (void)
{
	// and load our shaders
	R_CreateShaders (SurfSkyVS, SurfSkyPS, IDR_SURFSHADER);
}

void Sky_OnDestroy (void)
{
	SAFE_RELEASE (d3d_SkyCubemap);
}

void Sky_OnReset (void) {}
void Sky_OnLost (void) {}

void Sky_Register (void)
{
	static qboolean registered = false;

	if (!registered)
	{
		R_RegisterHandler (Sky_OnCreate, Sky_OnLost, Sky_OnReset, Sky_OnDestroy);
		registered = true;
	}
}


void R_DrawSkyChain (msurface_t *chain)
{
	surfbatch_t batch;

	// RenderState is set in the caller, not here, so that it's consistent with R_DrawWaterChain
	// stage 5 gets the sky
	R_SetCubeTexture (TS_SKY, d3d_SkyCubemap);

	R_SetVertexShader (d3d_SurfSkyVS);
	R_SetPixelShader (d3d_SurfSkyPS);

	R_BeginBatchingSurfaces (&batch);

	for (msurface_t *surf = chain; surf; surf = surf->texturechain)
		R_BatchSurface (&batch, surf, NULL);

	R_EndBatchingSurfaces (&batch);
}


void R_SetupSkyTransform (refdef_t *rd)
{
	glmatrix skymatrix;

	// this is z-going-up with rotate and flip to orient the sky correctly
	R_LoadMatrix (&skymatrix, 0, 0, 1, 0, -1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1);

	R_RotateMatrixAxis (&skymatrix, rd->time * skyrotate, skyaxis[0], skyaxis[2], skyaxis[1]);
	R_TranslateMatrix (&skymatrix, -rd->vieworigin[0], -rd->vieworigin[1], -rd->vieworigin[2]);

	R_SetVertexShaderConstant4fv (VSC_SKYMATRIX, skymatrix.m16, 4);
}


void R_SetSky (char *name, float rotate, vec3_t axis)
{
	if (!name || !name[0])
	{
		SAFE_RELEASE (d3d_SkyCubemap);
		skyname[0] = 0;
		return;
	}

	char pathname[MAX_QPATH];
	int maxextent = 0;
	int scaled_size = 0;

	char *suf[6] = {"ft", "bk", "up", "dn", "rt", "lf"};

	// if the name hasn't changed we can keep the old sky (provided we have one)
	if (!strcmp (skyname, name) && d3d_SkyCubemap) return;

	Q_strncpyz (skyname, name, sizeof (skyname));
	skyrotate = rotate;
	Vector3Copy (skyaxis, axis);

	// because we're building a cubemap here we must first load the individual textures, find the size we need for the cubemap,
	// then load the faces from those.  the individual textures can be discarded when done.
	for (int i = 0; i < 6; i++)
	{
		Com_sprintf (pathname, sizeof (pathname), "env/%s%s.tga", skyname, suf[i]);

		// the image is only needed while building the cubemap and can be released as an unused image when done
		// !!!!! note: can't use 0 for registration_sequence because then it wouldn't be freed at all !!!!!
		if ((sky_images[i] = R_FindImage (pathname, it_sky)) == r_notexture)
		{
			Com_sprintf (pathname, sizeof (pathname), "env/%s%s.pcx", skyname, suf[i]);

			if ((sky_images[i] = R_FindImage (pathname, it_sky)) == r_notexture)
				sky_images[i] = r_notexture;
			else sky_images[i]->registration_sequence = -1;
		}
		else sky_images[i]->registration_sequence = -1;

		// find the max size we're going to use for the cubemap (fixme: could this ever be 0?)
		D3DSURFACE_DESC desc;

		sky_images[i]->texture->lpVtbl->GetLevelDesc (sky_images[i]->texture, 0, &desc);

		if (desc.Width > maxextent) maxextent = desc.Width;
		if (desc.Height > maxextent) maxextent = desc.Height;
	}

	// this should never happen as we replace a sky face that failed to load with the grey texture
	if (!maxextent)
	{
		ri->Con_Printf (PRINT_ALL, "Failed to load sky %s\n", name);
		return;
	}

	if (!(d3d_DeviceCaps.TextureCaps & D3DPTEXTURECAPS_CUBEMAP_POW2))
	{
		// full unconditional non-power-of-two support
		scaled_size = maxextent;
	}
	else
	{
		// no non-power-of-two support
		// we don't support conditional because we need to use wrap modes for non-mipped textures too
		for (scaled_size = 1; scaled_size < maxextent; scaled_size <<= 1);
	}

	// the old cubemap is no longer valid
	SAFE_RELEASE (d3d_SkyCubemap);

	// now build the cubemap
	if (SUCCEEDED (d3d_Device->lpVtbl->CreateCubeTexture (d3d_Device, scaled_size, 1, 0, D3DFMT_X8R8G8B8, D3DPOOL_MANAGED, &d3d_SkyCubemap, NULL)))
	{
		// load the faces from the sky_images array to the cubemap
		for (int i = 0; i < 6; i++)
		{
			IDirect3DSurface9 *src = NULL;
			IDirect3DSurface9 *dst = NULL;

			// note workaround for excessive type-safety weenieness here
			if (SUCCEEDED (sky_images[i]->texture->lpVtbl->GetSurfaceLevel (sky_images[i]->texture, 0, &src)))
			{
				if (SUCCEEDED (d3d_SkyCubemap->lpVtbl->GetCubeMapSurface (d3d_SkyCubemap, (D3DCUBEMAP_FACES) ((int) D3DCUBEMAP_FACE_POSITIVE_X + i), 0, &dst)))
				{
					R_LoadSurfaceFromSurface (dst, src);
					dst->lpVtbl->Release (dst);
				}

				src->lpVtbl->Release (src);
			}
		}
	}
}


