/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "gl_local.h"
#include "resource.h"

extern void R_BuildLightMap (msurface_t *surf, image_t *lm);
extern void R_DirtyLightStyles (void);

static IDirect3DVertexDeclaration9 *d3d_SurfDecl = NULL;

static IDirect3DVertexShader9 *d3d_SurfLightmappedVS = NULL;
static IDirect3DPixelShader9 *d3d_SurfLightmappedPS = NULL;

static IDirect3DVertexShader9 *d3d_SurfDynamicVS = NULL;
static IDirect3DPixelShader9 *d3d_SurfDynamicPS = NULL;

static IDirect3DVertexShader9 *d3d_SurfAlphaVS = NULL;
static IDirect3DPixelShader9 *d3d_SurfAlphaPS = NULL;

static IDirect3DVertexBuffer9 *d3d_SurfaceVerts = NULL;
static IDirect3DIndexBuffer9 *d3d_SurfaceIndexes = NULL;

static int r_numsurfaceverts = 0;
static int r_firstsurfindex = 0;

void Surf_OnCreate (void)
{
	D3DVERTEXELEMENT9 layout[] = {
		VDECL (0, offsetof (brushpolyvert_t, xyz), D3DDECLTYPE_FLOAT3, D3DDECLUSAGE_POSITION, 0),
		VDECL (0, offsetof (brushpolyvert_t, st), D3DDECLTYPE_FLOAT2, D3DDECLUSAGE_TEXCOORD, 0),
		VDECL (0, offsetof (brushpolyvert_t, lm), D3DDECLTYPE_FLOAT2, D3DDECLUSAGE_TEXCOORD, 1),
		VDECL (0, offsetof (brushpolyvert_t, scroll), D3DDECLTYPE_FLOAT1, D3DDECLUSAGE_TEXCOORD, 2),
		VDECL (0, offsetof (brushpolyvert_t, styles), D3DDECLTYPE_UBYTE4, D3DDECLUSAGE_TEXCOORD, 3),
		VDECL (0, offsetof (brushpolyvert_t, normal), D3DDECLTYPE_FLOAT3, D3DDECLUSAGE_NORMAL, 0),
		D3DDECL_END ()
	};

	d3d_SurfDecl = R_CreateVertexDeclaration (layout);

	R_CreateShaders (SurfLightmappedVS, SurfLightmappedPS, IDR_SURFSHADER);
	R_CreateShaders (SurfDynamicVS, SurfDynamicPS, IDR_SURFSHADER);
	R_CreateShaders (SurfAlphaVS, SurfAlphaPS, IDR_SURFSHADER);
}


void Surf_OnReset (void)
{
	d3d_Device->lpVtbl->CreateIndexBuffer (
		d3d_Device,
		MAX_BATCHED_SURFINDEXES * sizeof (unsigned int),
		D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY,
		D3DFMT_INDEX32,
		D3DPOOL_DEFAULT,
		&d3d_SurfaceIndexes,
		NULL);

	r_firstsurfindex = 0;
	R_DirtyLightStyles ();
}


void Surf_OnLost (void)
{
	SAFE_RELEASE (d3d_SurfaceIndexes);
}


void Surf_OnDestroy (void)
{
	// release all lightmaps in any registration sequence
	R_ReleaseLightmaps (-1);
	SAFE_RELEASE (d3d_SurfaceVerts);
}


void Surf_Register (void)
{
	static qboolean registered = false;

	if (!registered)
	{
		R_RegisterHandler (Surf_OnCreate, Surf_OnLost, Surf_OnReset, Surf_OnDestroy);
		registered = true;
	}
}

msurface_t *r_skychain = NULL;
msurface_t *r_waterchain = NULL;

// this is sized at MAX_GLTEXTURES so we won't need to check it for overflow
image_t *r_frame_images[MAX_GLTEXTURES];
int r_num_frame_images;
int r_image_frame_count = 0;


void R_NewImageFrame (void)
{
	// this ensures that we only walk over images that are actually touched in this frame
	// it's split off from R_ClearTextureChains so that we can clear chains while still
	// keeping the current image list valid, which we need for dlights
	r_image_frame_count++;
	r_num_frame_images = 0;
}


void R_ClearTextureChains (void)
{
	for (int i = 0; i < numgltextures; i++)
		gltextures[i].texturechain = NULL;

	R_ClearLightmapChains ();

	r_skychain = NULL;
	r_waterchain = NULL;
}


void R_BeginBatchingSurfaces (surfbatch_t *batch)
{
	batch->lastimage = NULL;
	batch->numindexes = 0;
	batch->numsurfaces = 0;
}


void R_BatchSurface (surfbatch_t *batch, msurface_t *surf, image_t *image)
{
	// this can be called with NULL if we don't want this function to manage texture changes for us
	if (image != batch->lastimage)
	{
		// on the first texture change there will be no batched surfaces so this will return immediately
		R_EndBatchingSurfaces (batch);

		// only change texture if an image was specified
		if (image)
			R_SetTexture (TS_DIFFUSE, image);

		batch->lastimage = image;
	}

	if (batch->numsurfaces + 1 >= MAX_BATCHED_SURFACES)
		R_EndBatchingSurfaces (batch);

	if (batch->numindexes + surf->numindexes >= MAX_BATCHED_SURFINDEXES)
		R_EndBatchingSurfaces (batch);

	batch->surfaces[batch->numsurfaces] = surf;
	batch->numsurfaces++;
	batch->numindexes += surf->numindexes;
}


void R_EndBatchingSurfaces (surfbatch_t *batch)
{
	if (batch->numsurfaces && batch->numindexes)
	{
		// always draw batched even if there is only one because:
		// (1) we're doing multipass rendering, and:
		// (2) otherwise the same surf may be batched in one pass and unbatched in another, but:
		// (3) DrawPrimitive is not invariant with DrawIndexedPrimitive on some hardware.  different depths.
		// yes, this sucks
		DWORD iblock = D3DLOCK_NOOVERWRITE | D3DLOCK_NOSYSLOCK;

		if (r_firstsurfindex + batch->numindexes >= MAX_BATCHED_SURFINDEXES)
		{
			iblock = D3DLOCK_DISCARD | D3DLOCK_NOSYSLOCK;
			r_firstsurfindex = 0;
		}

		volatile unsigned int *ndx = NULL;
		int r_firstdipvert = 0x7fffffff;
		int r_lastdipvert = 0;

		if (SUCCEEDED (d3d_SurfaceIndexes->lpVtbl->Lock (d3d_SurfaceIndexes, r_firstsurfindex * sizeof (unsigned int), batch->numindexes * sizeof (unsigned int), (void **) &ndx, iblock)))
		{
			for (int i = 0; i < batch->numsurfaces; i++)
			{
				msurface_t *surf = batch->surfaces[i];

				for (int v = 2; v < surf->numedges; v++, ndx += 3)
				{
					ndx[0] = surf->firstvertex;
					ndx[1] = surf->firstvertex + (v - 1);
					ndx[2] = surf->firstvertex + v;
				}

				// fixup DIP params (this sucks too and thank christ it's gone in newer APIs)
				if (surf->firstvertex < r_firstdipvert) r_firstdipvert = surf->firstvertex;
				if (surf->lastvertex > r_lastdipvert) r_lastdipvert = surf->lastvertex;
			}

			d3d_SurfaceIndexes->lpVtbl->Unlock (d3d_SurfaceIndexes);
			d3d_Device->lpVtbl->DrawIndexedPrimitive (d3d_Device, D3DPT_TRIANGLELIST, 0, r_firstdipvert, r_lastdipvert, r_firstsurfindex, batch->numindexes / 3);

			c_draw_call++;
			r_firstsurfindex += batch->numindexes;

			// this is consistent with the original counts but because the texture chain is
			// taken in a single draw call it's no longer representative of performance
			c_brush_polys += batch->numsurfaces;
		}
	}

	// this can't use R_BeginBatchingSurfaces for the reset because it mustn't clear the texture
	batch->numindexes = 0;
	batch->numsurfaces = 0;
}


/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
===============
*/
image_t *R_TextureAnimation (mtexinfo_t *tex, int frame)
{
	int	c = frame % tex->numframes;

	if (tex->next)
	{
		while (c)
		{
			tex = tex->next;
			c--;
		}
	}

	return tex->image;
}


void R_DrawAlphaChain (msurface_t *chain)
{
	surfbatch_t batch;

	// RenderState is set in the caller, not here, so that we can reuse it for alpha surfaces
	R_SetVertexShader (d3d_SurfAlphaVS);
	R_SetPixelShader (d3d_SurfAlphaPS);

	R_BeginBatchingSurfaces (&batch);

	for (msurface_t *surf = chain; surf; surf = surf->texturechain)
		R_BatchSurface (&batch, surf, R_TextureAnimation (surf->texinfo, surf->animframe));

	R_EndBatchingSurfaces (&batch);
}


void R_DrawAlphaSurfaces (msurface_t *surfs[], int numsurfs)
{
	if (!numsurfs) return;

	// go back to the basic state
	R_SetRenderStates (RS_DEPTHTEST | RS_CULLCCW | RS_ALPHABLEND);

	// decl is common to all
	R_SetVertexDeclaration (d3d_SurfDecl);

	// verts and indexes are common to all
	R_SetVertexBuffer (0, d3d_SurfaceVerts, 0, sizeof (brushpolyvert_t));
	R_SetIndexBuffer (d3d_SurfaceIndexes);

	// if flags change we need to flush the current draw and begin a new one; starting with the
	// first surf so that we can always use "lastflags" for deciding which chain drawer to call
	// and ensuring that a draw with a NULL chain won't happen when we hit the first surface...
	int lastflags = surfs[0]->texinfo->flags;
	msurface_t *alphachain = NULL;

	// chaining will reverse the order so we walk through the list backwards here
	for (int i = numsurfs - 1; i >= 0; i--)
	{
		msurface_t *surf = surfs[i];

		// remove planeback from the comparison so they'll batch better
		if ((surf->texinfo->flags & ~SURF_PLANEBACK) != (lastflags & ~SURF_PLANEBACK))
		{
			// lastflags was given a value so this won't trigger on the first surf (and try to draw a NULL chain!)
			if (lastflags & SURF_WARP)
				R_DrawWaterChain (alphachain);
			else R_DrawAlphaChain (alphachain);

			// store back and clear the chain
			lastflags = surf->texinfo->flags;
			alphachain = NULL;
		}

		// chaining will reverse the order so we walked through the list backwards above
		surf->texturechain = alphachain;
		alphachain = surf;
	}

	// draw anything left over
	if (lastflags & SURF_WARP)
		R_DrawWaterChain (alphachain);
	else R_DrawAlphaChain (alphachain);
}


void R_DrawSolidChain (msurface_t *chain, image_t *image, const int pass)
{
	// unwind the chain for this pass
	int numsurfs = 0;

	for (msurface_t *surf = chain; surf; surf = surf->texturechain)
	{
		if (surf->texinfo->flags & SURF_WARP) continue;
		if (pass >= surf->numstyles) continue;

		R_BuildLightmapChain (surf, pass);
		numsurfs++;
	}

	if (!numsurfs) return;

	// pass 0 gets standard state, subsequent passes get additive blending
	if (pass == 0)
		R_SetRenderStates (RS_DEPTHTEST | RS_DEPTHWRITE | RS_CULLCCW);
	else R_SetRenderStates (RS_DEPTHTEST | RS_ADDITIVEBLEND);

	R_SetTexture (TS_DIFFUSE, image);

	R_SetVertexShader (d3d_SurfLightmappedVS);
	R_SetPixelShader (d3d_SurfLightmappedPS);

	// draw the lightmapped chain
	R_DrawLightmappedChain (pass);
}


void R_DrawDLightChains (dlight_t *dl)
{
	surfbatch_t batch;

	// all of our state is still up, so set some shaders and constants and proceed to draw it
	R_SetVertexShader (d3d_SurfDynamicVS);
	R_SetPixelShader (d3d_SurfDynamicPS);

	R_BeginBatchingSurfaces (&batch);

	// we could unbatch this and store it in it's own chain, meaning that dlights would no longer need to clear the chains
	// and the test is just if that chain is NULL or not - it would need to be a global in r_light but it could be static
	// (2015-05-13 - I have no idea what I was talking about there)
	for (int i = 0; i < r_num_frame_images; i++)
	{
		image_t *image = r_frame_images[i];

		if (!image->texture) continue;
		if (!image->registration_sequence) continue;
		if (!image->texturechain) continue;

		for (msurface_t *surf = image->texturechain; surf; surf = surf->texturechain)
			R_BatchSurface (&batch, surf, image);

		image->texturechain = NULL;
	}

	R_EndBatchingSurfaces (&batch);
}


/*
================
R_DrawTextureChains
================
*/
void R_DrawTextureChains (entity_t *e)
{
	// nothing to draw (can this ever happen???)
	// this test is INVALID because sky might be added which wouldn't bump the count, thereby causing it to
	// not be drawn.  it's also unnecessary because we're checking e->numdrawsurfs anyway
	// if (!r_num_frame_images) return;

	// this test is valid
	if (!e->numdrawsurfs) return;

	R_SetEntityTransform (&e->localmatrix, e->flags);

	// decl is common to all
	R_SetVertexDeclaration (d3d_SurfDecl);

	// verts and indexes are common to all
	R_SetVertexBuffer (0, d3d_SurfaceVerts, 0, sizeof (brushpolyvert_t));
	R_SetIndexBuffer (d3d_SurfaceIndexes);

	// draw solid chains in 4 passes, one for each lightstyle
	for (int pass = 0; pass < (r_fullbright->value ? 1 : 4); pass++)
	{
		// sending this per pass, NOT per texture.  sure, it'll be sent for some passes where there are no surfs.
		// so what?  it's only one register.  (this actually sucks and it would be cooler to be able to drop it
		// into a shader as a param.  even on D3D 10+ SV_InstanceID is 0-based irrespective of what first
		// instance you supply to the draw call; whoever in microsoft designed that must have been smoking the good stuff that day.
		R_SetVertexShaderConstant1f (VSC_SURFACEPASS, pass);

		for (int i = 0; i < r_num_frame_images; i++)
		{
			image_t *image = r_frame_images[i];

			if (!image->texture) continue;
			if (!image->registration_sequence) continue;
			if (!image->texturechain) continue;

			// draw the chains
			R_DrawSolidChain (image->texturechain, image, pass);
		}
	}

	// pick up water
	if (r_waterchain)
	{
		// go back to standard state - state isn't set in R_DrawWaterChain so that
		// we can reuse it for alpha surfaces
		R_SetRenderStates (RS_DEPTHTEST | RS_DEPTHWRITE | RS_CULLCCW);

		// and draw it
		R_DrawWaterChain (r_waterchain);
		r_waterchain = NULL;
	}

	// pick up sky
	if (r_skychain)
	{
		// go back to a default state with writing disabled (some effects in some maps depend on this)
		R_SetRenderStates (RS_DEPTHTEST | RS_CULLCCW);

		// and draw it
		R_DrawSkyChain (r_skychain);
		r_skychain = NULL;
	}

	// these ents should have dlights too
	e->flags |= RF_DYNAMICLIGHT;

	// add dynamic lights (e->model->firstnode is 0 for the world so this is valid)
	// done last because it clears the chains (R_PushDlights needs to check for e->model->firstnode < 0 which happens on boss1)
	R_PushDlights (e, e->model->nodes + e->model->firstnode);
}


void R_ChainSurface (entity_t *e, msurface_t *surf)
{
	// store out the frame for animations so that we can subsequently reuse it without having to carry the entity around everywhere
	surf->animframe = e->currframe;

	if (surf->texinfo->flags & SURF_SKY)
	{
		// add to the sky chain
		surf->texturechain = r_skychain;
		r_skychain = surf;
		e->numdrawsurfs++;
	}
	else if (surf->texinfo->flags & SURF_TRANS33)
		R_AddAlphaObject (e, surf, 0.333f);
	else if (surf->texinfo->flags & SURF_TRANS66)
		R_AddAlphaObject (e, surf, 0.666f);
	else if (e->flags & RF_TRANSLUCENT)
		R_AddAlphaObject (e, surf, 0.25f);
	else if (surf->texinfo->flags & SURF_WARP)
	{
		surf->texturechain = r_waterchain;
		r_waterchain = surf;
		e->numdrawsurfs++;
	}
	else
	{
		image_t *image = R_TextureAnimation (surf->texinfo, surf->animframe);

		// add to the list of visible images
		if (image->framecount != r_image_frame_count)
		{
			// this was sized at MAX_GLTEXTURES so we don't need to check it for overflow
			r_frame_images[r_num_frame_images] = image;
			r_num_frame_images++;

			// and mark as active
			image->framecount = r_image_frame_count;
		}

		// the polygon is visible, so add it to the appropriate texture sorted chain
		surf->texturechain = image->texturechain;
		image->texturechain = surf;

		// mark the surface as a candidate for dlights
		surf->dlightframe = r_dlightframecount;
		e->numdrawsurfs++;
	}
}


void R_BeginBuildingSurfaces (void)
{
	SAFE_RELEASE (d3d_SurfaceVerts);
	r_numsurfaceverts = 0;
}


void R_RegisterSurface (msurface_t *surf)
{
	// count the number of styles on the surf
	for (surf->numstyles = 0; surf->numstyles < MAXLIGHTMAPS && surf->styles[surf->numstyles] != 255; surf->numstyles++);

	// bound to range (is 0 styles valid?  shouldn't matter, it will just get a black lightmap texture
	if (surf->numstyles < 1) surf->numstyles = 1;
	if (surf->numstyles > 4) surf->numstyles = 4;

	// set up drawing params
	surf->firstvertex = r_numsurfaceverts;
	r_numsurfaceverts += surf->numedges;
	surf->lastvertex = r_numsurfaceverts - 1;

	surf->numindexes = (surf->numedges - 2) * 3;
}


void R_EndBuildingSurfaces (model_t *mod, dbsp_t *bsp)
{
	brushpolyvert_t *verts = NULL;

	d3d_Device->lpVtbl->CreateVertexBuffer (d3d_Device, sizeof (brushpolyvert_t) * r_numsurfaceverts, D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &d3d_SurfaceVerts, NULL);
	d3d_SurfaceVerts->lpVtbl->Lock (d3d_SurfaceVerts, 0, 0, (void **) &verts, 0);

	for (int i = 0; i < mod->numsurfaces; i++)
	{
		msurface_t *surf = &mod->surfaces[i];
		brushpolyvert_t *surfverts = &verts[surf->firstvertex];
		byte styles[4] = {0, 0, 0, 0};

		// set up styles
		for (int maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
			styles[maps] = surf->styles[maps];

		for (int v = 0; v < surf->numedges; v++, surfverts++)
		{
			int lindex = bsp->surfedges[surf->firstedge + v];
			float *vec = (lindex > 0) ? bsp->vertexes[bsp->edges[lindex].v[0]].point : bsp->vertexes[bsp->edges[-lindex].v[1]].point;

			Vector3Copy (surfverts->xyz, vec);

			if (surf->texinfo->flags & SURF_WARP)
			{
				// precalc as much of this as possible so that we don't need a more complex FS
				surfverts->st[0] = Vector3Dot (vec, surf->texinfo->vecs[0]) * 0.015625f;
				surfverts->st[1] = Vector3Dot (vec, surf->texinfo->vecs[1]) * 0.015625f;

				surfverts->lm[0] = Vector3Dot (vec, surf->texinfo->vecs[1]) * M_PI * 0.015625f;
				surfverts->lm[1] = Vector3Dot (vec, surf->texinfo->vecs[0]) * M_PI * 0.015625f;
			}
			else
			{
				// diffuse texture coordinates
				surfverts->st[0] = (Vector3Dot (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]) / surf->texinfo->image->width;
				surfverts->st[1] = (Vector3Dot (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]) / surf->texinfo->image->height;

				// lightmap texture coordinates
				R_SetLightmapTexCoords (surfverts->lm, surf, vec);
			}

			// copy over scroll factor
			if (surf->texinfo->flags & SURF_FLOWING)
				surfverts->scroll = 1.0f;
			else surfverts->scroll = 0.0f;

			// copy over styles
			*((unsigned *) surfverts->styles) = *((unsigned *) styles);

			// copy over normal, adjusting for planeback
			if (surf->flags & SURF_PLANEBACK)
				Vector3Scalef (surfverts->normal, surf->plane->normal, -1.0f);
			else Vector3Copy (surfverts->normal, surf->plane->normal);
		}

		// store the model used to create this surface
		surf->model = mod;
	}

	d3d_SurfaceVerts->lpVtbl->Unlock (d3d_SurfaceVerts);

	// for the next map
	r_numsurfaceverts = 0;
}


