/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// sys_win.h

#include "quakedef.h"

qboolean s_win95;

int			starttime;
int			ActiveApp;
qboolean Minimized;

static HANDLE		hinput, houtput;

unsigned	sys_msg_time;
unsigned	sys_frame_time;


static HANDLE		qwclsemaphore;

#define	MAX_NUM_ARGVS	128
int			argc;
char		*argv[MAX_NUM_ARGVS];


/*
===============================================================================

SYSTEM IO

===============================================================================
*/


void Sys_Error (char *error, ...)
{
	va_list		argptr;
	char		text[1024];

	CL_Shutdown ();
	Qcommon_Shutdown ();

	va_start (argptr, error);
	vsprintf (text, error, argptr);
	va_end (argptr);

	MessageBox (NULL, text, "Error", 0 /* MB_OK */);

	if (qwclsemaphore)
		CloseHandle (qwclsemaphore);

	// shut down QHOST hooks if necessary
	DeinitConProc ();

	exit (1);
}

void Sys_Quit (void)
{
	CL_Shutdown ();
	Qcommon_Shutdown ();
	CloseHandle (qwclsemaphore);

	if (dedicated && dedicated->value)
		FreeConsole ();

	// shut down QHOST hooks if necessary
	DeinitConProc ();

	exit (0);
}


void WinError (void)
{
	LPVOID lpMsgBuf;

	FormatMessage (
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
		NULL,
		GetLastError (),
		MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		(LPTSTR) &lpMsgBuf,
		0,
		NULL
		);

	// Display the string.
	MessageBox (NULL, lpMsgBuf, "GetLastError", MB_OK | MB_ICONINFORMATION);

	// Free the buffer.
	LocalFree (lpMsgBuf);
}

//================================================================


/*
================
Sys_Init
================
*/
void Sys_Init (void)
{
	OSVERSIONINFO	vinfo;

	vinfo.dwOSVersionInfoSize = sizeof (vinfo);

	if (!GetVersionEx (&vinfo))
		Sys_Error ("Couldn't get OS info");

	if (vinfo.dwMajorVersion < 4)
		Sys_Error ("Quake2 requires windows version 4 or greater");

	if (vinfo.dwPlatformId == VER_PLATFORM_WIN32s)
		Sys_Error ("Quake2 doesn't run on Win32s");
	else if (vinfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
		s_win95 = true;

	if (dedicated->value)
	{
		if (!AllocConsole ())
			Sys_Error ("Couldn't create dedicated server console");

		hinput = GetStdHandle (STD_INPUT_HANDLE);
		houtput = GetStdHandle (STD_OUTPUT_HANDLE);

		// let QHOST hook in
		InitConProc (argc, argv);
	}
}


static char	console_text[256];
static int	console_textlen;

/*
================
Sys_ConsoleInput
================
*/
char *Sys_ConsoleInput (void)
{
	INPUT_RECORD	recs[1024];
	int		dummy;
	int		ch, numread, numevents;

	if (!dedicated || !dedicated->value)
		return NULL;

	for (;;)
	{
		if (!GetNumberOfConsoleInputEvents (hinput, &numevents))
			Sys_Error ("Error getting # of console events");

		if (numevents <= 0)
			break;

		if (!ReadConsoleInput (hinput, recs, 1, &numread))
			Sys_Error ("Error reading console input");

		if (numread != 1)
			Sys_Error ("Couldn't read console input");

		if (recs[0].EventType == KEY_EVENT)
		{
			if (!recs[0].Event.KeyEvent.bKeyDown)
			{
				ch = recs[0].Event.KeyEvent.uChar.AsciiChar;

				switch (ch)
				{
				case '\r':
					WriteFile (houtput, "\r\n", 2, &dummy, NULL);

					if (console_textlen)
					{
						console_text[console_textlen] = 0;
						console_textlen = 0;
						return console_text;
					}

					break;

				case '\b':
					if (console_textlen)
					{
						console_textlen--;
						WriteFile (houtput, "\b \b", 3, &dummy, NULL);
					}

					break;

				default:
					if (ch >= ' ')
					{
						if (console_textlen < sizeof (console_text) - 2)
						{
							WriteFile (houtput, &ch, 1, &dummy, NULL);
							console_text[console_textlen] = ch;
							console_textlen++;
						}
					}

					break;
				}
			}
		}
	}

	return NULL;
}


/*
================
Sys_ConsoleOutput

Print text to the dedicated console
================
*/
void Sys_ConsoleOutput (char *string)
{
	int		dummy;
	char	text[256];

	if (!dedicated || !dedicated->value)
		return;

	if (console_textlen)
	{
		text[0] = '\r';
		memset (&text[1], ' ', console_textlen);
		text[console_textlen + 1] = '\r';
		text[console_textlen + 2] = 0;
		WriteFile (houtput, text, console_textlen + 2, &dummy, NULL);
	}

	WriteFile (houtput, string, strlen (string), &dummy, NULL);

	if (console_textlen)
		WriteFile (houtput, console_text, console_textlen, &dummy, NULL);
}


/*
================
Sys_SendKeyEvents

Send Key_Event calls
================
*/
void Sys_SendKeyEvents (void)
{
	MSG msg;

	while (PeekMessage (&msg, NULL, 0, 0, PM_NOREMOVE))
	{
		if (!GetMessage (&msg, NULL, 0, 0))
			Sys_Quit ();

		sys_msg_time = msg.time;
		TranslateMessage (&msg);
		DispatchMessage (&msg);
	}

	// grab frame time
	sys_frame_time = Sys_Milliseconds ();	// FIXME: should this be at start?
}


/*
================
Sys_GetClipboardData

================
*/
char *Sys_GetClipboardData (void)
{
	char *data = NULL;
	char *cliptext = NULL;

	if (OpenClipboard (NULL) != 0)
	{
		HANDLE hClipboardData;

		if ((hClipboardData = GetClipboardData (CF_TEXT)) != 0)
		{
			if ((cliptext = GlobalLock (hClipboardData)) != 0)
			{
				data = Z_Alloc (GlobalSize (hClipboardData) + 1);
				strcpy (data, cliptext);
				GlobalUnlock (hClipboardData);
			}
		}

		CloseClipboard ();
	}

	return data;
}


void Sys_FreeClipboardData (void *data)
{
	Z_Free (data);
}


/*
==============================================================================

WINDOWS CRAP

==============================================================================
*/

/*
=================
Sys_AppActivate
=================
*/
void Sys_AppActivate (void)
{
	ShowWindow (cl_hwnd, SW_RESTORE);
}


/*
========================================================================

GAME DLL

========================================================================
*/

static HINSTANCE	game_library;

/*
=================
Sys_UnloadGame
=================
*/
void Sys_UnloadGame (void)
{
	if (!FreeLibrary (game_library))
		Com_Error (ERR_FATAL, "FreeLibrary failed for game library");

	game_library = NULL;
}

/*
=================
Sys_GetGameAPI

Loads the game dll
=================
*/
void *Sys_GetGameAPI (void *parms)
{
	void *(*GetGameAPI) (void *);
	char	name[MAX_OSPATH];
	char	*path;
	char	cwd[MAX_OSPATH];
	const char *gamename = "gamex86.dll";

#ifdef NDEBUG
	const char *debugdir = "release";
#else
	const char *debugdir = "debug";
#endif

	if (game_library)
		Com_Error (ERR_FATAL, "Sys_GetGameAPI without Sys_UnloadingGame");

	// check the current debug directory first for development purposes
	_getcwd (cwd, sizeof (cwd));
	Com_sprintf (name, sizeof (name), "%s/%s/%s", cwd, debugdir, gamename);
	game_library = LoadLibrary (name);

	if (game_library)
	{
		Com_DPrintf ("LoadLibrary (%s)\n", name);
	}
	else
	{
		// check the current directory for other development purposes
		Com_sprintf (name, sizeof (name), "%s/%s", cwd, gamename);
		game_library = LoadLibrary (name);

		if (game_library)
		{
			Com_DPrintf ("LoadLibrary (%s)\n", name);
		}
		else
		{
			// now run through the search paths
			path = NULL;

			while (1)
			{
				path = FS_NextPath (path);

				if (!path)
					return NULL;		// couldn't find one anywhere

				Com_sprintf (name, sizeof (name), "%s/%s", path, gamename);
				game_library = LoadLibrary (name);

				if (game_library)
				{
					Com_DPrintf ("LoadLibrary (%s)\n", name);
					break;
				}
			}
		}
	}

	GetGameAPI = (void *) GetProcAddress (game_library, "GetGameAPI");

	if (!GetGameAPI)
	{
		Sys_UnloadGame ();
		return NULL;
	}

	return GetGameAPI (parms);
}

//=======================================================================


/*
==================
Sys_ParseCommandLine

==================
*/
void Sys_ParseCommandLine (LPSTR lpCmdLine)
{
	argc = 1;
	argv[0] = "exe";

	while (*lpCmdLine && (argc < MAX_NUM_ARGVS))
	{
		while (*lpCmdLine && ((*lpCmdLine <= 32) || (*lpCmdLine > 126)))
			lpCmdLine++;

		if (*lpCmdLine)
		{
			argv[argc] = lpCmdLine;
			argc++;

			while (*lpCmdLine && ((*lpCmdLine > 32) && (*lpCmdLine <= 126)))
				lpCmdLine++;

			if (*lpCmdLine)
			{
				*lpCmdLine = 0;
				lpCmdLine++;
			}
		}
	}
}


/*
==================
WinMain

==================
*/
int Sys_WaitForFrame (int oldtime)
{
	if (!(cl_timedemo->value && cls.state == ca_active))
	{
		// not in a timedemo so ensure that at least 1ms has elapsed before running a frame
		while (Sys_Milliseconds () - oldtime < 1)
			YieldProcessor ();
	}

	// in a timedemo
	return Sys_Milliseconds ();
}


// http://ntcoder.com/bab/tag/getuserprofiledirectory/
#include "userenv.h"
#pragma comment (lib, "userenv.lib")

char *Sys_GetUserHomeDir (void)
{
	static char szHomeDirBuf[MAX_PATH] = {0};

	// We need a process with query permission set
	HANDLE hToken = 0;
	DWORD BufSize = MAX_PATH;

	OpenProcessToken (GetCurrentProcess (), TOKEN_QUERY, &hToken);

	// Returns a path like C:/Documents and Settings/nibu if my user name is nibu
	GetUserProfileDirectory (hToken, szHomeDirBuf, &BufSize);

	// Close handle opened via OpenProcessToken
	CloseHandle (hToken);

	return szHomeDirBuf;
}


BOOL Sys_DirectoryExists (LPCTSTR szPath)
{
	DWORD dwAttrib = GetFileAttributes (szPath);
	return (dwAttrib != INVALID_FILE_ATTRIBUTES && (dwAttrib & FILE_ATTRIBUTE_DIRECTORY));
}


BOOL Sys_FileExists (LPCTSTR szPath)
{
	DWORD dwAttrib = GetFileAttributes (szPath);
	return (dwAttrib != INVALID_FILE_ATTRIBUTES && !(dwAttrib & FILE_ATTRIBUTE_DIRECTORY));
}


void Sys_SetWorkingDirectory (void)
{
	int i, j, k;

	char *homeDir = Sys_GetUserHomeDir ();

	const char *testDirs[] = {
		// these are the locations I have Quake installed to on various PCs; you may want to change it yourself
		homeDir,
		"\\Desktop Crap",
		"\\Games",
		"",
		NULL
	};

	const char *qDirs[] = {
		"Quake2",
		"Quake 2",
		"Quake II",
		"QuakeII",
		"Q2",
		"QII",
		NULL
	};

	const char *qFiles[] = {
		"pak0.pak",
		"gamex86.dll",
		"config.cfg",
		NULL
	};

	// try to find baseq2 content in the test paths
	for (i = 0;; i++)
	{
		if (!testDirs[i]) break;
		if (!Sys_DirectoryExists (testDirs[i])) continue;

		for (j = 0;; j++)
		{
			if (!qDirs[j]) break;
			if (!Sys_DirectoryExists (va ("%s\\%s", testDirs[i], qDirs[j]))) continue;

			for (k = 0;; k++)
			{
				if (!qFiles[k]) break;
				if (!Sys_FileExists (va ("%s\\%s\\"BASEDIRNAME"\\%s", testDirs[i], qDirs[j], qFiles[k]))) continue;

				SetCurrentDirectory (va ("%s\\%s", testDirs[i], qDirs[j]));
				return;
			}
		}
	}
}


int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG	msg;
	int	oldtime, newtime;

#ifdef _DEBUG
	Sys_SetWorkingDirectory ();
#endif

	Sys_ParseCommandLine (lpCmdLine);

	Qcommon_Init (argc, argv);
	oldtime = Sys_Milliseconds ();

	/* main window message loop */
	while (1)
	{
		// release any temp allocations before the frame
		Scratch_FreeToLowMark (0);

		if (Minimized || (dedicated && dedicated->value))
		{
			// if at a full screen console, don't update unless needed
			Sleep (1);
		}

		while (PeekMessage (&msg, NULL, 0, 0, PM_NOREMOVE))
		{
			if (!GetMessage (&msg, NULL, 0, 0))
				Com_Quit ();

			sys_msg_time = msg.time;
			TranslateMessage (&msg);
			DispatchMessage (&msg);
		}

		newtime = Sys_Milliseconds ();
		Qcommon_Frame (newtime - oldtime);
		oldtime = newtime;

		// not in a running timedemo so yield some CPU (to do: cvar-ize this)
		if (!(cl_timedemo->value && cls.state == ca_active))
			Sleep (1);
	}

	// never gets here
	return TRUE;
}

